"""
  TODO: 
    - basic file management:
        - add file history recording for all ops
        - add own file deletion `diskapi_remove`
  
"""
import shutil

import pandas as pd
import cv2
import os
import zipfile
import yaml
from naeural_core import constants as ct
from typing import List, Union, Tuple
from naeural_core.local_libraries.vision.ffmpeg_writer import FFmpegWriter

from naeural_core.ipfs import R1FSEngine

def assert_folder(folder : str):
  assert folder in ['data', 'models', 'output']

class _DiskAPIMixin(object):

  def __init__(self):
    super(_DiskAPIMixin, self).__init__()
    return

  # Utils section
  if True:
    def is_path_safe(self, path, base_path=None):
      """
      Method for checking if a certain path is safe(it's inside the base_path).
      If the base_path is not provided, the base_path will be the base folder of the cache.
      Parameters
      ----------
      path - string, path to be checked
      base_path - string, the base path to check against.
        If not provided, the base path will be the base folder of the cache.

      Returns
      -------
      bool, True if the path is safe, False otherwise
      """
      if base_path is None:
        base_path = self.log.get_base_folder()
      abs_base_directory = os.path.abspath(base_path)

      # This conversion is done to avoid issues with both symbolic links and "../" segments in the path.
      real_path_base = os.path.realpath(abs_base_directory)
      real_path = os.path.realpath(path)

      return os.path.commonpath([real_path_base, real_path]) == real_path_base
  # endif Utils

  # Dataframe serialization section
  if True:
    def _diskapi_save_dataframe(self, df : pd.DataFrame, filename : str, folder : str,
                                ignore_index : bool = True, compress : bool = False, mode : str = 'w',
                                header : Union[bool, List[str]] = True,
                                also_markdown : bool = False, verbose : bool = True,
                                as_parquet : bool = False):

      """
      Parameters:
      -----------
      df: pandas.DataFrame, mandatory
        The dataframe that should be saved

      filename: str, mandatory
        Relative path to `folder` (in local cache), where the dataframe should be saved

      folder: str, mandatory
        The folder in local cache
        Possible values: 'data', 'output', 'models'

      ignore_index: bool, optional
        Decides whether to ignore or not the index when writing
        The default value is True

      compress: bool, optional
        Save to zipped pickle
        The default value is False.

      mode: str, optional
        The writing mode in csv.
        Possible values: 'w' - write; 'a' - append
        The default value is 'w' - write.

      header: bool or List[str], optional
        Write out the column names. If a list of strings is given it is assumed to be aliases for the column names.
        This may be set to False for 'append' mode, for all but not the first save call.
        The default value is False.

      verbose: bool, optional
        Controls logging when saving the dataframe
        The default value is True
        
      as_parquet: bool, optional
        Save as parquet file. Default is False

      Returns:
      -------
      str
        full path to the saved file
      """

      mode = mode.lower()
      assert mode in ['w', 'a']
      assert_folder(folder)

      _, full_path = self.log.save_dataframe(
        df=df, fn=filename, folder=folder,
        ignore_index=ignore_index, compress=compress,
        mode=mode, header=header, also_markdown=also_markdown, verbose=verbose,
        as_parquet=as_parquet,
      )

      return full_path


    def diskapi_save_dataframe_to_data(self, df : pd.DataFrame, filename : str,
                                       ignore_index : bool = True, compress : bool = False, mode : str = 'w',
                                       header : Union[bool, List[str]] = True,
                                       also_markdown : bool = False, verbose : bool = True,
                                       as_parquet : bool = False
                                       ):
      """
      Shortcut to _diskapi_save_dataframe.
      """
      return self._diskapi_save_dataframe(
        df=df, filename=filename, folder='data',
        ignore_index=ignore_index, compress=compress, mode=mode, header=header,
        also_markdown=also_markdown, verbose=verbose, as_parquet=as_parquet,
      )

    def diskapi_save_dataframe_to_models(self, df : pd.DataFrame, filename : str,
                                         ignore_index : bool = True, compress : bool = False, mode : str = 'w',
                                         header : Union[bool, List[str]] = True,
                                         also_markdown : bool = False, verbose : bool = True,
                                         as_parquet : bool = False,
                                         ):
      """
      Shortcut to _diskapi_save_dataframe.
      """
      return self._diskapi_save_dataframe(
        df=df, filename=filename, folder='models',
        ignore_index=ignore_index, compress=compress, mode=mode, header=header,
        also_markdown=also_markdown, verbose=verbose, as_parquet=as_parquet,
      )

    def diskapi_save_dataframe_to_output(self, df : pd.DataFrame, filename : str,
                                         ignore_index : bool = True, compress : bool = False, mode : str = 'w',
                                         header : Union[bool, List[str]] = True,
                                         also_markdown : bool = False, verbose : bool = True,
                                         as_parquet : bool = False,
                                         ):
      """
      Shortcut to _diskapi_save_dataframe.
      """
      return self._diskapi_save_dataframe(
        df=df, filename=filename, folder='output',
        ignore_index=ignore_index, compress=compress, mode=mode, header=header,
        also_markdown=also_markdown, verbose=verbose, as_parquet=as_parquet,
      )

    def _diskapi_load_dataframe(self, filename : str, folder : str,
                                decompress : bool = False, timestamps : Union[str, List[str]] = None):
      """
      Parameters:
      ----------
      filename: str, mandatory
        Relative path to `folder` (in local cache), from where the dataframe should be loaded
        If filename ends in ".zip" then the loading will also uncompress in-memory

      folder: str, mandatory
        The folder in local cache
        Possible values: 'data', 'output', 'models'

      decompress: bool, optional
        Should be True if the file was saved with `compress=True`
        The default value is False.

      timestamps: str or List[str], optional
        Column names that should be parsed as dates when loading the dataframe
        The default is None

      Returns:
      --------
      pandas.DataFrame
      """
      assert_folder(folder)
      df = self.log.load_dataframe(
        fn=filename, 
        folder=folder, 
        decompress=decompress, 
        timestamps=timestamps
      )
      return df

    def diskapi_load_dataframe_from_data(self, filename : str, decompress : bool = False,
                                         timestamps : Union[str, List[str]] = None):
      """
      Shortcut to _diskapi_load_dataframe.
      """
      return self._diskapi_load_dataframe(
        filename=filename, folder='data', decompress=decompress, timestamps=timestamps
      )

    def diskapi_load_dataframe_from_models(self, filename : str, decompress : bool = False,
                                           timestamps : Union[str, List[str]] = None):
      """
      Shortcut to _diskapi_load_dataframe.
      """
      return self._diskapi_load_dataframe(
        filename=filename, folder='models', decompress=decompress, timestamps=timestamps)

    def diskapi_load_dataframe_from_output(self, filename : str, decompress : bool = False,
                                           timestamps : Union[str, List[str]] = None):
      """
      Shortcut to _diskapi_load_dataframe.
      """
      return self._diskapi_load_dataframe(
        filename=filename, folder='output', decompress=decompress, timestamps=timestamps
      )
  # endif

  # Pickle serialization section
  if True:
    def _diskapi_save_pickle(
        self, obj: object, filename: str, folder: str,
        subfolder: str = None, compress: bool = False,
        verbose: bool = True
    ):
      """
      Parameters:
      -----------
      obj: object, mandatory
        Any object that can be pickled to be saved

      filename: str, mandatory
        Relative path to `folder` (in local cache), where the object should be pickled

      folder: str, mandatory
        The folder in local cache
        Possible values: 'data', 'output', 'models'

      subfolder: str, optional
        A subfolder in `folder` where the object should be pickled

      compress: bool, optional
        if compression is required. It is activated also by setting '.pklz' as extension for `filename`
        The default value is False

      verbose: bool, optional
        Controls logging when saving the object
        The default value is True

      Returns:
      --------
      str
        full_path to the saved object
      """
      assert_folder(folder)

      full_path = self.log.save_pickle(
        data=obj, fn=filename, folder=folder,
        subfolder_path=subfolder,
        compressed=compress, verbose=verbose,
        locking=True
      )

      return full_path

    def diskapi_save_pickle_to_data(
        self, obj: object, filename: str, subfolder: str = None,
        compress : bool = False, verbose : bool = True
    ):
      """
      Shortcut to _diskapi_save_pickle.
      """
      return self._diskapi_save_pickle(
        obj=obj, filename=filename, subfolder=subfolder,
        folder='data', compress=compress, verbose=verbose
      )

    def diskapi_save_pickle_to_models(
        self, obj: object, filename: str, subfolder: str = None,
        compress: bool = False, verbose: bool = True
    ):
      """
      Shortcut to _diskapi_save_pickle.
      """
      return self._diskapi_save_pickle(
        obj=obj, filename=filename, subfolder=subfolder,
        folder='models', compress=compress, verbose=verbose
      )

    def diskapi_save_pickle_to_output(
        self, obj: object, filename: str, subfolder: str = None,
        compress: bool = False, verbose: bool = True
    ):
      """
      Shortcut to _diskapi_save_pickle.
      """
      return self._diskapi_save_pickle(
        obj=obj, filename=filename, subfolder=subfolder,
        folder='output', compress=compress, verbose=verbose
      )

    def _diskapi_load_pickle(
        self, filename: str, folder: str, subfolder: str = None,
        decompress: bool = False, verbose: bool = True
    ):
      """
      Parameters:
      -----------
      filename: str, mandatory
        Relative path to `folder` (in local cache), from where a pickled object should be loaded
        If filename ends in ".pklz" then the loading will also decompress

      folder: str, mandatory
        The folder in local cache
        Possible values: 'data', 'output', 'models'

      decompress: bool, optional
        Should be True if the file was saved with `compress=True`
        The default value is False.

      verbose: bool, optional
        Controls logging when loading the object
        The default value is True

      Returns:
      --------
      object
      """
      assert_folder(folder)

      obj = self.log.load_pickle(
        fn=filename, folder=folder, subfolder_path=subfolder,
        decompress=decompress, verbose=verbose,
        locking=True
      )

      return obj

    def diskapi_load_pickle_from_data(
        self, filename : str, subfolder: str = None,
        decompress: bool = False, verbose: bool = True
    ):
      """
      Shortcut to _diskapi_load_pickle.
      """
      return self._diskapi_load_pickle(
        filename=filename, folder='data', subfolder=subfolder,
        decompress=decompress, verbose=verbose
      )

    def diskapi_load_pickle_from_models(
        self, filename: str, subfolder: str = None,
        decompress: bool = False, verbose: bool = True
    ):
      """
      Shortcut to _diskapi_load_pickle.
      """
      return self._diskapi_load_pickle(
        filename=filename, folder='models', subfolder=subfolder,
        decompress=decompress, verbose=verbose
      )

    def diskapi_load_pickle_from_output(
        self, filename: str, subfolder: str = None,
        decompress: bool = False, verbose: bool = True
    ):
      """
      Shortcut to _diskapi_load_pickle.
      """
      return self._diskapi_load_pickle(
        filename=filename, folder='output', subfolder=subfolder,
        decompress=decompress, verbose=verbose
      )
  # endif PICKLE

  # JSON serialization section
  if True:
    def _diskapi_save_json(self, dct, filename : str, folder : str, indent : bool = True):
      """
      Parameters:
      -----------
      dct: dict or list, mandatory
        A dictionary or list like object that can be saved as json

      filename: str, mandatory
        Relative path to `folder` (in local cache), where the dictionary should be saved

      folder: str, mandatory
        The folder in local cache
        Possible values: 'data', 'output', 'models'

      indent: bool, optional
        Flag that controls the indentation in the output file
        The default value is False

      Returns:
      --------
      str
        full_path to the saved json
      """
      assert_folder(folder)

      full_path = self.log.thread_safe_save(
        datafile=filename,
        data_json=dct,
        folder=folder,
        indent=indent,
        locking=True
      )

      return full_path

    def diskapi_save_json_to_data(self, dct, filename : str, indent : bool = True):
      """
      Shortcut to _diskapi_save_json
      """
      return self._diskapi_save_json(dct=dct, filename=filename, folder='data', indent=indent)

    def diskapi_save_json_to_models(self, dct, filename : str, indent : bool = True):
      """
      Shortcut to _diskapi_save_json
      """
      return self._diskapi_save_json(dct=dct, filename=filename, folder='models', indent=indent)

    def diskapi_save_json_to_output(self, dct, filename : str, indent : bool = True):
      """
      Shortcut to _diskapi_save_json
      """
      return self._diskapi_save_json(dct=dct, filename=filename, folder='output', indent=indent)

    def _diskapi_load_json(
      self, 
      filename : str, 
      folder : str, 
      subfolder : str = None,
      verbose : bool = True
    ):
      """
      Parameters:
      -----------
      filename: str, mandatory
        Relative path to `folder` (in local cache), from where a json should be loaded

      folder: str, mandatory
        The folder in local cache
        Possible values: 'data', 'output', 'models'

      verbose: bool, optional
        Controls logging when loading the json
        The default value is True

      Returns:
      --------
      dict
      """
      assert_folder(folder)

      dct = self.log.load_json(
        fname=filename, folder=folder, subfolder=subfolder, numeric_keys=True,
        locking=True, verbose=verbose
      )

      return dct
    
    
    def diskapi_load_json(self, fn, verbose : bool = True):
      """
      Loads a json file from the local cache in one of the three folders: data, models, output.
      
      Parameters
      -----------
      
      fn: str, mandatory
        The filename of the json file to be loaded. The file should include the extension '.json'
        as well as the path to one of the three folders: data, models, output.
      """
      
      is_ext =  isinstance(fn, str) and fn.endswith('.json')
      if not is_ext:
        self.P(f"JSON load: Invalid filename: {fn}", color='r')
        return None
      output_folder = self.log.get_output_folder()
      models_folder = self.log.get_models_folder()
      data_folder = self.log.get_data_folder()
      found = False      
      for folder in [output_folder, models_folder, data_folder]:
        if folder in fn:
          found = True
      if not found:
        self.P(f"JSON load: Invalid filename location for {fn}", color='r')
        return None
      
      dct = self.log.load_json(
        fname=fn, folder=None, subfolder=None, numeric_keys=True,
        locking=True, verbose=verbose
      )
      return dct
    
    def diskapi_load_yaml(self, fn, verbose : bool = True):
      """
      Loads a yaml file from the local cache in one of the three folders: data, models, output.
      
      Parameters
      -----------
      
      fn: str, mandatory
        The filename of the yaml file to be loaded. The file should include the extension '.yaml'
        as well as the path to one of the three folders: data, models, output.
      """
      is_ext = isinstance(fn, str) and (fn.endswith('.yaml') or fn.endswith('.yml'))
      if not is_ext: 
        self.P(f"Invalid filename: {fn}", color='r')
        return None      
      
      output_folder = self.log.get_output_folder()
      models_folder = self.log.get_models_folder()
      data_folder = self.log.get_data_folder()
      found = False      
      for folder in [output_folder, models_folder, data_folder]:
        if folder in fn:
          found = True
      if not found:
        self.P(f"Invalid filename location for {fn}", color='r')
        return None
      if not os.path.isfile(fn):
        self.P(f"File not found: {fn}", color='r')
        return None
      try:
        dct = None
        with open(fn, 'rt') as fd:
          if verbose:
            self.P("Loading yaml file: {}".format(fn))
          dct = yaml.load(fd, Loader=yaml.FullLoader)
      except Exception as e:
        self.P(f"Failed to load yaml file: {fn}: {e}", color='r')
      return dct
    

    def diskapi_load_json_from_data(self, filename : str, verbose : bool = True):
      """
      Shortcut to _diskapi_load_json.
      """
      return self._diskapi_load_json(filename=filename, folder='data', verbose=verbose)

    def diskapi_load_json_from_models(self, filename : str, verbose : bool = True):
      """
      Shortcut to _diskapi_load_json.
      """
      return self._diskapi_load_json(filename=filename, folder='models', verbose=verbose)

    def diskapi_load_json_from_output(self, filename : str, verbose : bool = True):
      """
      Shortcut to _diskapi_load_json.
      """
      return self._diskapi_load_json(filename=filename, folder='output', verbose=verbose)
  # endif

  # Video serialization section
  if True:
    def _diskapi_create_video_file(self, filename : str, folder : str, fps : int, str_codec : str,
                                   frame_size : Tuple[int,int], universal_codec : str = 'XVID'):
      """
      Parameters:
      -----------
      filename: str, mandatory
        Relative path to `folder` (in local cache), where the video file should be create

      folder: str, mandatory
        The folder in local cache
        Possible values: 'data', 'output', 'models'

      fps : int, mandatory
        fps of the video

      str_codec: str, mandatory
        which codec to be used
        ["H264", "h264", "mp4v", "XVID" , "MJPG", "DIVX", "FMP4"]

      frame_size: Tuple[int, int], mandatory
        H, W of each frame in the video

      universal_codec: str, optional
        codec to fallback on it if the chosen `str_codec` is not installed
        applies for all codecs except "H264"!
        The default value is "XVID"

      Returns:
      --------
      _:
        a handler to the video that can be used in `diskapi_write_video_frame`

      str:
        full_path to the video file
      """
      assert_folder(folder)
      video_path = os.path.join(self.log.get_target_folder(folder), filename)
      os.makedirs(os.path.split(video_path)[0], exist_ok=True)

      handler = None
      if str_codec.upper() == 'H264':
        self.P("Creating '{}' ffmpeg output movie '...{}' using video FPS: {}".format(
          str_codec, video_path[-20:], fps),
          color='y'
        )
        handler = FFmpegWriter(
          filename=video_path,
          fps=fps,
          frameSize=frame_size,
          log=self.log,
        )
      else:
        self.P("Creating cv2 output movie '...{}' using video FPS: {} and codec '{}'".format(
          video_path[-20:], fps, str_codec),
          color='y'
        )
        fourcc = cv2.VideoWriter_fourcc(*str_codec)

        handler = cv2.VideoWriter(
          filename=video_path,
          fourcc=fourcc,
          fps=fps,
          frameSize=frame_size,
        )

        if not handler.isOpened():
          msg = "Failed to create output movie '{}' using codec '{}'".format(video_path, str_codec)
          self._create_notification(
            notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
            msg=msg,
          )
          self.P(msg, color='r')
          self.P("  Creating output movie '...{}' using video codec '{}'".format(video_path[-20:], universal_codec), color='m')
          fourcc = cv2.VideoWriter_fourcc(*universal_codec)
          handler = cv2.VideoWriter(
            filename=video_path,
            fourcc=fourcc,
            fps=fps,
            frameSize=frame_size,
          )
          # if still doesn't work then...
          if not handler.isOpened():
            msg = "CRITICAL: Failed to create output movie '{}' using codec '{}'".format(video_path, universal_codec)
            raise ValueError(msg)
          # end critical error
        # end fail with given codec
      # end ffmpeg-vs-cv2

      if handler.isOpened():
        self.P("  Opened video handler for '{}' with codec '{}'".format(video_path, str_codec), color='g')

      return handler, video_path


    def diskapi_create_video_file_to_data(self, filename : str, fps : int, str_codec : str,
                                          frame_size : Tuple[int,int], universal_codec : str = 'XVID'):
      """
      Shortcut to `_diskapi_create_video_file`
      """
      return self._diskapi_create_video_file(
        filename=filename, folder='data',
        fps=fps, str_codec=str_codec,
        frame_size=frame_size, universal_codec=universal_codec
      )

    def diskapi_create_video_file_to_models(self, filename : str, fps : int, str_codec : str,
                                            frame_size : Tuple[int,int], universal_codec : str = 'XVID'):
      """
      Shortcut to `_diskapi_create_video_file`
      """
      return self._diskapi_create_video_file(
        filename=filename, folder='models',
        fps=fps, str_codec=str_codec,
        frame_size=frame_size, universal_codec=universal_codec
      )

    def diskapi_create_video_file_to_output(self, filename : str, fps : int, str_codec : str,
                                            frame_size : Tuple[int,int], universal_codec : str = 'XVID'):
      """
      Shortcut to `_diskapi_create_video_file`
      """
      return self._diskapi_create_video_file(
        filename=filename, folder='output',
        fps=fps, str_codec=str_codec,
        frame_size=frame_size, universal_codec=universal_codec
      )

    def diskapi_write_video_frame(self, handler, frame):
      """
      Parameters:
      -----------
      handler: _, mandatory
        the handler returned by `diskapi_create_video_file`

      frame: np.ndarray, mandatory
        the frame to be written in the video file.
        Must have the the same H,W specified in `diskapi_create_video_file`
      """
      handler.write(frame)
      return
  # endif

  # Dataset save section
  if True:
    def _get_file_path(self, folder: str, subdir: str, filename: str, extension: str):
      """
      Method for getting the full path to a file in local cache.
      Parameters
      ----------
      folder - string, the folder in local cache
      subdir - string, the subfolder in local cache
      filename - string, the name of the file
      extension - string, the extension of the file

      Returns
      -------
      string, the full path to the file
      """
      save_dir = self.log.get_target_folder(folder)
      if subdir is not None and len(subdir) > 0:
        save_dir = os.path.join(save_dir, subdir)
      if extension is not None and len(extension) > 0:
        if not filename.endswith(f'.{extension}'):
          filename = f'{filename}.{extension}'
      # endif extension is not None and len(extension) > 0
      return os.path.join(save_dir, filename)

    def _diskapi_save_image(self, image, filename: str, extension: str, folder: str, subdir: str):
      """
      Method for saving an image to local cache.
      Parameters
      ----------
      image - np.ndarray, the image to be saved
      filename - string, the name of the file
      extension - string, the extension of the file
      folder - string, the folder in local cache
      subdir - string, the subfolder in local cache

      Returns
      -------
      bool, True if the image was saved successfully, False otherwise
      """
      assert_folder(folder)
      save_path = self._get_file_path(
        folder=folder,
        subdir=subdir,
        filename=filename,
        extension=extension
      )
      os.makedirs(os.path.dirname(save_path), exist_ok=True)
      res = cv2.imwrite(img=image[:, :, ::-1], filename=save_path)  # TODO: use PIL instead of opencv
      if res:
        result = save_path
      else:
        result = None
      return result
      

    def diskapi_load_image(self, filename: str, folder: str, subdir: str):
      """
      Method for loading an image from local cache.
      Parameters
      ----------
      filename - string, the name of the file
      folder - string, the folder in local cache
      subdir - string, the subfolder in local cache

      Returns
      -------
      np.ndarray, the loaded image
      """
      assert_folder(folder)
      load_path = self._get_file_path(
        folder=folder,
        subdir=subdir,
        filename=filename,
        extension=''
      )
      if not os.path.exists(load_path):
        self.P(f'File {load_path} does not exist')
        return None
      return cv2.imread(load_path)[:, :, ::-1]  # TODO: use PIL instead of opencv

    def diskapi_save_image_output(self, image, filename: str, subdir: str, extension: str = 'jpg'):
      """
      Shortcut to _diskapi_save_image.
      
      Parameters
      ----------
      image - np.ndarray, the image to be saved
      filename - string, the name of the file
      subdir - string, the subfolder in local cache
      extension - string, the extension of the file

      Returns
      -------
      bool, True if the image was saved successfully, False otherwise
      """
      return self._diskapi_save_image(
        image=image,
        filename=filename,
        folder='output',
        subdir=subdir,
        extension=extension
      )


    def _diskapi_save_file(
      self, 
      data: any, 
      filename: str, 
      folder: str, 
      subdir: str, 
      extension: str,
      save_bytes: bool = False,
    ):
      """
      Method for saving a file to local cache.
      
      Parameters
      ----------
      data - string or list, the data to be saved
      
      filename - string, the name of the file
      
      folder - string, the folder in local cache
      
      subdir - string, the subfolder in local cache
      
      extension - string, the extension of the file

      Returns
      -------
        str, the full path to the saved file or None if the file was not saved
      
      
      """
      save_path = None
      try:
        assert_folder(folder)
        save_path = self._get_file_path(
          folder=folder,
          subdir=subdir,
          filename=filename,
          extension=extension
        )
        if isinstance(data, list):
          data = '\n'.join([x for x in data])
        os.makedirs(os.path.dirname(save_path), exist_ok=True)
        mode = 'w' if not save_bytes else 'wb'
        with open(save_path, mode) as out:
          out.write(data)
      except Exception as e:
        self.P(f'Failed to save file {filename} to {folder} with error: {e}')
        save_path = None
      return save_path

    def diskapi_save_file_to_output(self, data: any, filename: str, subdir: str = '', extension: str = ''):
      """
      Shortcut to _diskapi_save_file for ouput folder.
      
      Parameters
      ----------
      data - string or list, the data to be saved
      
      filename - string, the name of the file
      
      subdir - string, the subfolder in local cache
      
      extension - string, the extension of the file

      Returns
      -------
      
        str: the full path to the saved file or None if the file was not saved
      
      """
      return self._diskapi_save_file(
        data=data,
        filename=filename,
        folder='output',
        subdir=subdir,
        extension=extension
      )
  # endif

  # Zipfile section
  if True:
    def diskapi_zip_dir(self, dir_path, zip_path=None, include_dir=True):
      """
      Zip the contents of an entire folder (with that folder included).
      Parameters
      ----------
      dir_path - string, path of directory to zip
      zip_path - string, path of the output zip file. If None, zip_path will be dir_path + ".zip"
      include_dir - bool, whether to include the directory itself in the zip file

      Returns
      -------
      string, the path to the zipped directory
      """
      if zip_path is None:
        zip_path = dir_path + '.zip'
      self.P(f'Archiving {dir_path} to {zip_path}')
      ziph = zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED)
      for root, dirs, files in os.walk(dir_path):
        for file in files:
          file_root = os.path.join(dir_path, '..') if include_dir else dir_path
          ziph.write(os.path.join(root, file), os.path.relpath(os.path.join(root, file), file_root))
        # endfor files
      # endfor os.walk
      ziph.close()
      return zip_path

    def diskapi_unzip_dir(self, zip_path, dir_path=None):
      """
      Unzip a file into a given directory.
      Parameters
      ----------
      zip_path - string, path to .zip file
      dir_path - string, path to directory into which to unzip the input .zip file

      Returns
      -------
      string, the path to the unzipped directory
      """
      if dir_path is None:
        dir_path, ext = os.path.splitext(zip_path)
      self.P(f'Unzipping {zip_path} to {dir_path}')
      with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        zip_ref.extractall(dir_path)
      return dir_path
  # endif

  # Delete files/directories section
  if True:
    def diskapi_delete_file(self, file_path):
      """
      Delete a file from disk if safe.
      Parameters
      ----------
      file_path - string, path to the file to be deleted

      Returns
      -------

      """
      if self.is_path_safe(file_path):
        try:
          self.P(f'Trying to delete safe file {file_path}')
          os.remove(file_path)
          self.P(f'Deleted safe file from {file_path}')
        except Exception as e:
          self.P(e)
        # end try-except
      else:
        self.P(f'Provided unsafe file path ({file_path}) for deletion. Nothing happened.')
      # endif path is safe
      return

    def diskapi_delete_directory(self, dir_path):
      """
      Delete a directory from disk if safe.
      Parameters
      ----------
      dir_path - string, path to the directory to be deleted

      Returns
      -------

      """
      if self.is_path_safe(dir_path):
        try:
          self.P(f'Trying to delete safe directory {dir_path}')
          shutil.rmtree(dir_path)
          self.P(f'Deleted safe directory from {dir_path}')
        except Exception as e:
          self.P(e)
        # end try-except
      else:
        self.P(f'Provided unsafe directory path ({dir_path}) for deletion. Nothing happened.')
      # endif path is safe
      return
  # endif

  # Copy files section
  if True:
    def diskapi_copy_file(self, src_path, dst_path):
      """
      Copy a file from src to dst if safe.
      Parameters
      ----------
      src_path - string, path to the source file
      dst_path - string, path to the destination file

      Returns
      -------

      """
      if self.is_path_safe(src_path) and self.is_path_safe(dst_path):
        try:
          self.P(f'Trying to copy safe file from {src_path} to {dst_path}')
          dst_dir = os.path.dirname(dst_path)
          os.makedirs(dst_dir, exist_ok=True)
          shutil.copy2(src_path, dst_path)
          self.P(f'Copied safe file from {src_path} to {dst_path}')
        except Exception as e:
          self.P(e)
        # end try-except
      else:
        self.P(f'Provided unsafe file paths for copying. Nothing happened.')
      # endif paths are safe
      return
  # endif
  
  # binary objects handling
  if True:
    def diskapi_save_bytes_to_output(
      self, 
      data: bytes, 
      filename: str, 
      subdir: str = '', 
      extension: str = '', 
      from_base64: bool = False,
      from_base64_decompress: bool = False,
      from_base64_url_safe: bool = False,
    ):
      """
      Save bytes to a file in the output folder.
      
      
      Parameters
      ----------
      
      data - bytes, the data to be saved
      
      filename - string, the name of the file
      
      subdir - string, the subfolder in local cache
      
      extension - string, the extension of the file

      Returns
      -------
      
        str, the full path to the saved file or None if the file was not saved
        
      """
      if isinstance(data, str) and from_base64:
        data = self.base64_to_bytes(
          b64=data, decompress=from_base64_decompress, url_safe=from_base64_url_safe
        )
      return self._diskapi_save_file(
        data=data,
        filename=filename,
        folder='output',
        subdir=subdir,
        extension=extension,
        save_bytes=True,
      )


    def diskapi_load_r1fs_file(
      self, 
      filename: str, 
      verbose: bool = True,
      to_base64: bool = False,
      url_safe: bool = False,
      compress: bool = False,
    ):
      """
      Load a file from the R1FS folder.
      
      
      Parameters
      ----------
      
      filename - string, the name of the file
      
      
      verbose - bool, whether to print messages or not

      Returns
      -------
        bytes, the loaded file if not to_base64, str if to_base64
      """
      eng : R1FSEngine = self.r1fs
      target = eng.download_folder
      assert self.is_path_safe(filename), f"Filename {filename} is not safe"
      assert filename.startswith(target), f"Filename {filename} should start with {target}"
      if verbose:
        self.P(f"Loading binary content from {filename}")
      with open(filename, 'rb') as f:
        data = f.read()
      if to_base64:
        data = self.bytes_to_base64(data, url_safe=url_safe, compress=compress)
      return data



import gc
import numpy as np
import torch as th
import torchvision as tv
from naeural_core.local_libraries.nn.utils import get_dropout_rate

__VER__ = '0.2.0.0'


def clear_cache():
  gc.collect()
  th.cuda.empty_cache()
  return


def Pr(s=''):
  print('\r' + str(s), end='', flush=True)


def get_dropout(
    dropout=0.1,
    dropout_type='classic',
    value_type='constant',
    step=0,
    max_step=0
):
  """

  Parameters
  ----------
  dropout : float, The base dropout rate
  dropout_type : str, The type of dropout to be used
    if 'classic' the dropout used will be the standard dropout
    if 'spatial' the dropout used will be the spatial dropout
  value_type : str, The type of the dropout rate variation
    if 'constant' the dropout rate will be constant,
    if 'last' the dropout rate will be 0 until the last step,
    if 'progressive' the dropout rate will increase linearly from 0 to the specified dropout rate.
    otherwise the dropout rate will be 0
  step : int, The current step in the training process
  max_step : int, The maximum number of steps in the training process

  Returns
  -------

  """
  dropout_rate = get_dropout_rate(value_type, dropout, step, max_step)
  dropout_class = th.nn.Dropout2d if dropout_type == 'spatial' else th.nn.Dropout
  return dropout_class(dropout_rate)


def _th_normalize(th_x, sub_val, div_val, half=False):
  if half:
    th_x = th_x.half()
  else:
    th_x = th_x.float()
  if isinstance(sub_val, th.Tensor) or sub_val != 0:
    th_x -= sub_val
  th_x /= div_val
  return th_x


def auto_normalize(th_x, scale_min, scale_max):
  th_proc = (th_x - th_x.min()) / (th_x.max() - th_x.min())
  th_proc = th_proc * (scale_max - scale_min) + scale_min
  th_x = th.where(th_x.max() > 1, th_proc, th_x)
  return th_x


def test_norm(test=0):
  img = np.random.randint(0,255, size=(1000,1000,3), dtype='uint8')
  th_x = th.tensor(img, device=th.device('cuda'))
  if test == 0:
    th_out = _th_normalize(th_x, 0, 255, False)
  else:
    th_out = th_x / 255
  return th_out


def recompute_relative_y(old_size, new_size, old_y):
  """
  Method used to compute relative position after image was resized with `th_resize_with_pad`

  Parameters:
  old_size: tuple(int, int)
    The shape of the image before it was resized with `th_resize_with_pad`

  new_size: tuple(int, int)
    The shape of the image after it was resized with `th_resize_with_pad`

  old_y: tuple(float[0,1], float[0,1])
    Relative a point described relative position of the image before it was resized with `th_resize_with_pad`

  Returns:
  new_y: tuple(float[0,1], float[0,1])
    Relative a point described relative position of the image after it was resized with `th_resize_with_pad`

  -------------
  PROOF:

  old_ratio = old image width / height or height / width
  new_ratio = resized image width / height or height / width
  ratio_black_total = the ratio of the borders over the total image over a given dimension
  ratio_colored_total = 1 - ratio_black_total

  For visualisation we can imagine the following resized image (`*` = border):
       ____________________
      | ****************** |
      |                    |
      |                    |
      |                    |
      | ****************** |
       ---------------------
  We have
     ratio_black_total = 2 / 5 = 0.4
     ratio_color_total = 1 - ratio_black_total = 0.6
  -------

  new_ratio = new_width / new_height
  old_ratio = old_width / old_height

  as we have border along a single dimension:
  new_width = old_width + border_width
  new_height = old_height

  new_ratio - old_ratio = new_width / new_height - old_width / old_height
                        = (old_width +  border_width) / old_height - old_width / old_height
                        = border_width / old_height
  =>
  ratio_black_total = new_ratio - old_ratio

  (if the ratio is ratio_black_total it means wee need the take the inverses of `new_ratio` and `old_ratio`)


  new_coord = old_coord * ratio_colored_total + ratio_black_total / 2
             ( we need to find the point on    ( and add the top border )
              the colored part of the image )

  """
  if isinstance(old_y, np.ndarray):
    new_y = old_y.copy()
  elif isinstance(old_y, th.Tensor):
    new_y = th.clone(old_y)
  elif isinstance(old_y, list):
    new_y = old_y.copy()
  else:
    raise ValueError("Method `recompute_relative_y` not implemented for old_y type '{}'".format(type(old_y)))
  #endif
  old_ratio = old_size[0] / old_size[1]
  new_ratio = new_size[0] / new_size[1]

  ratio_black_total = new_ratio - old_ratio

  if ratio_black_total < 0:
    old_ratio = old_size[0] / old_size[1]
    new_ratio = new_size[0] / new_size[1]
    ratio_black_total = new_ratio - old_ratio
    ratio_colored_total = 1 - ratio_black_total

    new_y[0] = old_y[0] * ratio_colored_total + ratio_black_total / 2
    new_y[2] = old_y[2] * ratio_colored_total + ratio_black_total / 2
  else:
    ratio_colored_total = 1 - ratio_black_total

    new_y[1] = old_y[1] * ratio_colored_total + ratio_black_total / 2
    new_y[3] = old_y[3] * ratio_colored_total + ratio_black_total / 2
  return new_y


def inverse_recompute_relative_y(old_size, new_size, new_y):
  """
  Method used to compute the original relative position after image was resized with `th_resize_with_pad`

  Parameters:
  old_size: tuple(int, int)
    The shape of the image before it was resized with `th_resize_with_pad`

  new_size: tuple(int, int)
    The shape of the image after it was resized with `th_resize_with_pad`

  new_y: tuple(float[0,1], float[0,1])
    Relative a point described relative position of the image after it was resized with `th_resize_with_pad`


  Returns:
  old_y: tuple(float[0,1], float[0,1])
    Relative a point described relative position of the image before it was resized with `th_resize_with_pad`

  -------------
  PROOF:
  Following the `recompute_relative_y` proof we arrive at the conclusion that
      new_coord = old_coord * ratio_colored_total + ratio_black_total / 2

  So we can extract the old_coord from the above equation and result:
      old_coord = (new_coord - ratio_black_total / 2) / ratio_colored_total

  """
  if isinstance(new_y, np.ndarray):
    old_y = new_y.copy()
  elif isinstance(new_y, th.Tensor):
    old_y = th.clone(new_y)
  elif isinstance(new_y, list):
    old_y = new_y.copy()
  else:
    raise ValueError("Method `recompute_relative_y` not implemented for new_y type '{}'".format(type(new_y)))
  # endif

  old_ratio = old_size[0] / old_size[1]
  new_ratio = new_size[0] / new_size[1]

  ratio_black_total = new_ratio - old_ratio

  if ratio_black_total < 0:
    old_ratio = old_size[0] / old_size[1]
    new_ratio = new_size[0] / new_size[1]
    ratio_black_total = new_ratio - old_ratio
    ratio_colored_total = 1 - ratio_black_total

    old_y[0] = (new_y[0] - ratio_black_total / 2) / ratio_colored_total
    old_y[2] = (new_y[2] - ratio_black_total / 2) / ratio_colored_total
  else:
    ratio_colored_total = 1 - ratio_black_total

    old_y[1] = (new_y[1] - ratio_black_total / 2) / ratio_colored_total
    old_y[3] = (new_y[3] - ratio_black_total / 2) / ratio_colored_total
  return old_y

def th_resize_with_pad(img, h, w, 
                       device=None, 
                       pad_post_normalize=False,
                       normalize=True,
                       sub_val=0, div_val=255,
                       half=False,
                       return_original=False,
                       half_original=False,
                       normalize_original=False,
                       normalize_callback=None,
                       fill_value=None,
                       return_pad=False,
                       shrink_factor=None,
                       **kwargs
                       ):

  def _th_resize(th_img, normalize_callback=None, fill_value=None):
    # prep default callback
    if normalize_callback is None:
      normalize_callback = _th_normalize
    # if numpy load it to device before any other op
    if isinstance(th_img, np.ndarray):
      th_img = th.tensor(th_img, device=device)
    # check the shape
    if th_img.shape[-1] == 3:
      if len(th_img.shape) == 3:
        th_img = th_img.permute((2,0,1))
      elif len(th_img.shape) == 4:
        th_img = th_img.permute((0, 3, 1, 2))

    # copy  original
    th_img_orig = th_img * 1
    if fill_value is None:
      # fill with gray
      fill_value = 114 if th_img.dtype == th.uint8 else 114/255
    new_shape = h, w
    shape = th_img.shape[-2:]  # current shape [height, width] 
    # next code is quite tensor/graph safe
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if shrink_factor is not None:
      r /= shrink_factor
    new_unpad = int(round(shape[0] * r)), int(round(shape[1] * r))
    dw, dh = new_shape[1] - new_unpad[1], new_shape[0] - new_unpad[0]  # wh padding
    dw /= 2  # divide padding into 2 sides
    dh /= 2
    if shape != new_unpad:
      th_resized = tv.transforms.Resize(new_unpad)(th_img)
    else:
      th_resized = th_img
    # finally we have the padding params
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))

    if not pad_post_normalize:
      # usually pad before normalize
      th_resized_padded = tv.transforms.Pad((left, top, right, bottom), fill=fill_value)(th_resized)
    else:
      th_resized_padded = th_resized
    #endif

    if th_resized_padded.dtype == th.uint8 and normalize:
      th_resized_padded = normalize_callback(
        th_x=th_resized_padded,
        sub_val=sub_val, 
        div_val=div_val,
        half=half
        )
    #endif

    if pad_post_normalize:
      # pad after normalize if needed ...
      fill_value = 114/255
      th_resized_padded = tv.transforms.Pad((left, top, right, bottom), fill=fill_value)(th_resized_padded)
    #endif
    
    if len(th_img.shape) == 4:
      # send downstream the paddings - copy redundant but easier for downstream
      pad = [(top, left, bottom, right)] * th_img.shape[0]
    else:
      pad = [(top, left, bottom, right)]
    
    result = (th_resized_padded,)
    if return_original:
      if normalize_original:
        th_img_orig = _th_normalize(
          th_x=th_img_orig,
          sub_val=sub_val,
          div_val=div_val,
          half=half
        )
      if half_original:
        th_img_orig = th_img_orig.half()
      # endif
      result += (th_img_orig,)
    if return_pad:
      result += (pad,)
      
    return result[0] if len(result) == 1 else result

  
  if isinstance(img, list):
    images = []
    originals = []
    lst_original_shapes = []
    lst_pads = []
    for im in img:
      if isinstance(im, th.Tensor):
        shape = im.shape[1:]
      else:
        shape = im.shape[:-1]
      lst_original_shapes.append(shape)
      out = _th_resize(im, fill_value=fill_value, normalize_callback=normalize_callback)      
      orig, pads = None, None      
      if isinstance(out, tuple) and len(out) == 3:
        th_img, orig, pads = out
      elif isinstance(out, tuple) and len(out) == 2:
        if return_pad:
          th_img, pads = out
        else:
          th_img, orig = out        
      else:
        th_img = out
      # if tuples or just a tensor
      images.append(th_img.unsqueeze(0))
      if orig is not None:
        originals.append(orig)
      if pads is not None:
        lst_pads += pads # TLBR      
    # end for each variabile size image
    th_out = th.cat(images)
    th_originals = originals if return_original else None
  else:
    if isinstance(img, th.Tensor):
      lst_original_shapes = [x.shape[-2:] for x in img] if len(img.shape) == 4 else [img.shape[-2:]]
    elif isinstance(img, np.ndarray): # could be a NHWC numpy array
      lst_original_shapes = [x.shape[:-1] for x in img] if len(img.shape) == 4 else [img.shape[:-1]]
    else:
      raise ValueError("Uknown img input for th_resize_with_pad: {}".format(type(img)))
    out = _th_resize(img, fill_value=fill_value, normalize_callback=normalize_callback)
    th_out = out
    if isinstance(out, tuple):
      if len(out) == 3:
        th_out, th_originals, lst_pads = out
      elif len(out) == 2:
        if return_pad:
          th_out, lst_pads = out
        else:
          th_out, th_originals = out
      # endif len(out) == 2
    # endif isinstance(out, tuple)
  
  final_result = (th_out, lst_original_shapes)
  if return_original:
    final_result += (th_originals,)
  if return_pad:
    final_result += (lst_pads,)
  return final_result
  


def get_torch_n_params(model):
  pp = 0
  for p in list(model.parameters()):
    nn = 1
    for s in list(p.size()):
      nn = nn * s
    pp += nn
  return pp


def get_activation(act):
  if act.lower() == 'relu6':
    return th.nn.ReLU6()
  if act.lower() == 'relu':
    return th.nn.ReLU()
  elif act.lower() == 'tanh':
    return th.nn.Tanh()
  elif act.lower() == 'selu':
    return th.nn.SELU()
  elif act.lower() == 'gelu':
    return th.nn.GELU()
  elif act.lower() == 'sigmoid':
    return th.nn.Sigmoid()
  elif act.lower() == 'softmax':
    return th.nn.Softmax()
  else:
    raise ValueError("Unknown activation function '{}'".format(act))


def get_optimizer_class(opt):
  if opt.lower() == 'adam':
    return th.optim.Adam
  elif opt.lower() == 'sgd':
    return th.optim.SGD
  elif opt.lower() == 'adagrad':
    return th.optim.Adagrad
  elif opt.lower() == 'adadelta':
    return th.optim.Adadelta
  elif opt.lower() == 'rmsprop':
    return th.optim.RMSprop
  else:
    raise NotImplementedError('optimizer {} not implemented in `libraries.nn.th.utils.get_optimizer_class`'.format(opt))


def get_loss(loss):
  from naeural_core.local_libraries.nn.th.losses import ConstrativeLoss, HuberLoss, QuantileLoss
  if loss.lower() == 'mse':
    return th.nn.MSELoss
  elif loss.lower() == 'kl':
    return th.nn.KLDivLoss
  elif loss.lower() == 'bce':
    return th.nn.BCELoss
  elif loss.lower() == 'ce' or loss.lower() == 'crossentropy':
    return th.nn.CrossEntropyLoss
  elif loss.lower() == 'l1':
    return th.nn.L1Loss
  elif loss.lower() == 'contrastive':
    return ConstrativeLoss
  elif loss.lower() == 'huber':
    return HuberLoss
  elif loss.lower() == 'quantile':
    return QuantileLoss
  else:
    raise NotImplementedError("Unknown loss {}".format(loss))


def enforce_reproducibility(seed=1234):
  # Sets seed manually for both CPU and CUDA
  th.manual_seed(seed)
  # For atomic operations there is currently 
  # no simple way to enforce determinism, as
  # the order of parallel operations is not known.
  #
  # CUDNN
  th.backends.cudnn.deterministic = True
  th.backends.cudnn.benchmark = False
  # System based
  np.random.seed(seed)


def auto_regression_step_level(X, model, nr_steps, output_tensor_idx, keep_output_idx=None):
  """
  :param model:
  :param X: List of torch tensors
  :param nr_steps:
  :param output_tensor_idx:
  :return:
  """

  for step in range(nr_steps):
    if step < nr_steps - 1:
      x_input = [x[:, :-(nr_steps - step - 1)] for x in X]
    else:
      x_input = X

    preds = model(x_input)
    if keep_output_idx is not None:
      preds = preds[keep_output_idx]

    if step < nr_steps - 1:
      X[output_tensor_idx][:, -(nr_steps - step - 1)] = preds[:, -1]

  return preds[:, -nr_steps:].cpu().detach().numpy()


def transfer_weights_tf_to_th(tf_model, th_model):
  import tensorflow as tf
  weight_dict = dict()
  for layer in tf_model.layers:
    if type(layer) is tf.keras.layers.Conv1D:
      weight_dict[layer.get_config()['name'] + '.conv.weight'] = np.transpose(layer.get_weights()[0], (2, 1, 0))
      weight_dict[layer.get_config()['name'] + '.conv.bias'] = layer.get_weights()[1]
    elif type(layer) is tf.keras.layers.Dense:
      weight_dict[layer.get_config()['name'] + '.weight'] = np.transpose(layer.get_weights()[0], (1,0))
      weight_dict[layer.get_config()['name'] + '.bias'] = layer.get_weights()[1]
    elif type(layer) is tf.keras.layers.BatchNormalization:
      weight_dict[layer.get_config()['name'] + '.weight'] = layer.get_weights()[0]
      weight_dict[layer.get_config()['name'] + '.bias'] = layer.get_weights()[1]
    elif type(layer) is tf.keras.layers.Embedding:
      ### TODO: CHECK
      weight_dict[layer.get_config()['name']] = layer.get_weights()[0]

  pyt_state_dict = th_model.state_dict()
  parsed_keys = []
  for th_key in pyt_state_dict.keys():
    for tf_key in weight_dict.keys():
      if tf_key in th_key:
        pyt_state_dict[th_key] = th.from_numpy(weight_dict[tf_key])
        parsed_keys.append(tf_key)
      #endif
    #endfor
  #endfor
  weights_transferred = 0
  for weights in weight_dict.values():
    weights_transferred += weights.size
  assert weights_transferred == tf_model.count_params()
  print(len(weight_dict))
  print(len(parsed_keys))
  print(set)
  assert len(weight_dict) == len(parsed_keys), 'Transfer weights failed, layers do not match'
 # assert len(parsed_keys) == len(pyt_state_dict), 'Transfer weights failed, layers do not match'
  th_model.load_state_dict(pyt_state_dict)
  return th_model

def l2distance(x, y):
  return th.sqrt(((x-y)*(x-y)).clip(min=th.finfo(th.float32).eps))


if __name__ == "__main__":
  class Dummy(th.nn.Module):
    def __init__(self):
      super().__init__()

    def forward(self, x):
      return x[0] + x[1]

  model = Dummy()
  device = th.device("cuda" if th.cuda.is_available() else "cpu")

  input_tensors = [np.array([np.arange(10)] * 2) for x in range(2)]
  input_tensors[1] += 1
  input_tensors[0] *= 10
  input_tensors[0][1] += 100
  input_tensors[1][1] += 100


  nr_steps = 3
  input_tensors[0][:, -(nr_steps - 1):] = 0

  Y = auto_regression_step_level(
    model=model,
    X=input_tensors,
    nr_steps=3,
    output_tensor_idx=0,
    device=device
  )


  




import numpy as np
import torch as th
import torchvision.transforms as T
import torchvision.transforms.functional as F

from naeural_core.local_libraries.nn.th.utils import th_resize_with_pad

class PreprocessBrightness:
  def __init__(self, max_delta):
    self.transform = T.ColorJitter(brightness=max_delta)
    return

  def __call__(self, img):
    return self.transform(img)

class PreprocessRotation:
  def __init__(self, max_degrees):
    if max_degrees > 20:
      print("WARNING! In rotation preproc, |max_degrees| is greater than 20. Please be sure that you want higher rotation")
    self.transform = T.RandomRotation(max_degrees)
    return

  def __call__(self, img):
    return self.transform(img)

class PreprocessCropBottom:
  def __init__(self, max_prc, min_prc):
    self.max_prc = max_prc
    self.min_prc = min_prc
    return

  def __call__(self, img):
    shape = img.shape
    crop_prc = th.rand([]) * (self.max_prc - self.min_prc) + self.min_prc

    cropped_img = F.crop(img, 0, 0, shape[-2] - (crop_prc*shape[-2]).type(th.int32), shape[-1])

    return th_resize_with_pad(cropped_img, shape[-2], shape[-1], normalize=False)[0]

class PreprocessMinMaxNorm:
  def __init__(self, min_val, max_val, target_min_val, target_max_val):
    self.min_val = min_val
    self.max_val = max_val
    self.target_min_val = target_min_val
    self.target_max_val = target_max_val

  def __call__(self, th_x, **kwargs):
    th_x = (th_x - self.min_val) / (self.max_val - self.min_val)
    th_x = th_x * (self.target_max_val - self.target_min_val) + self.target_min_val

    return th_x


class PreprocessResizeWithPad:
  def __init__(self, h, w, normalize, normalize_callback=None, **kwargs):
    self.h = h
    self.w = w
    self.normalize = normalize
    self.normalize_callback = normalize_callback
    self.kwargs = kwargs

  def __call__(self, img, return_original_size=False):
    img, original_size = th_resize_with_pad(img, self.h, self.w, normalize=self.normalize, normalize_callback=self.normalize_callback, **self.kwargs)
    if return_original_size:
      th_original_size = th.tensor(original_size).to(img.device)
      return img, th_original_size
    else:
      return img


class NumpyPreprocessRandomCrop:
  def __init__(self, max_prc, min_prc):
    self.max_prc = max_prc
    self.min_prc = min_prc

  def __call__(self, img):
    shape = img.shape

    crop_prc_t = np.random.uniform(self.min_prc, self.max_prc)
    crop_prc_l = np.random.uniform(self.min_prc, self.max_prc)
    crop_prc_b = np.random.uniform(self.min_prc, self.max_prc)
    crop_prc_r = np.random.uniform(self.min_prc, self.max_prc)

    cropped_img = img[
      np.int(crop_prc_t * shape[0]):np.int((1- crop_prc_b) * shape[1]),
      np.int(crop_prc_l * shape[0]):np.int((1 - crop_prc_r) * shape[1])
    ]


    return cropped_img


PREPROCESS_TO_STR = {
  PreprocessResizeWithPad: 'PreprocessResizeWithPad',
  PreprocessMinMaxNorm: 'PreprocessMinMaxNorm',
  T.Resize: 'Resize'
}
STR_TO_PREPROCESS = {v: k for k, v in PREPROCESS_TO_STR.items()}


def preprocess_to_str(preprocess):
  """
  Method used to convert a preprocess object to a string
  Parameters
  ----------
  preprocess : object, The preprocess object to be converted to a string

  Returns
  -------
  str, The string representation of the preprocess object
  """
  return PREPROCESS_TO_STR.get(preprocess, None)


def str_to_preprocess(s=''):
  """
  Method used to convert a string to a preprocess object
  Parameters
  ----------
  s : str, The string representation of the preprocess object

  Returns
  -------
  object, The preprocess object
  """
  return STR_TO_PREPROCESS.get(s, None)




import threading
import psutil
from time import sleep
from naeural_core import Logger

from naeural_core.utils.sys_mon import SystemMonitor
                  

if __name__ == '__main__':
  l = Logger('THR', base_folder='.', app_folder='_cache')
  mon = SystemMonitor(
    log=l, 
    monitored_prefix='S_', 
    name='S_SysMon',
    max_size=20,
    tick=0.1,
    DEBUG=True # shows debug status dict
  )
  mon.start()

  mon.log_status()

  thread_data = {
    1 : {
      'flag' : False,
      'thread' : None,
    },
    2 : {
      'flag' : False,
      'thread' : None,
    },
  }

  def stop_thread(thr):
    thread_data[thr]['flag'] = True
    thread_data[thr]['thread'].join()
    return
  
  def start_thread(thr):
    thread_data[thr]['thread'].start()
    return
    
      
  
  def t1():
    cnt = 0
    while True:
      cnt += 1
      if thread_data[1]['flag']:
        l.P("Forcing quit {}".format(thread_data[1]['thread']), color='d')
        break
      else:
        sleep(0.5)
        l.P("Tick {} from {}".format(cnt, thread_data[1]['thread']), color='d')
    return  


  def t2():
    cnt = 0
    while True:
      cnt += 1
      if thread_data[2]['flag']:
        l.P("Forcing quit {}".format(thread_data[2]['thread']), color='d')
        break
      else:
        sleep(1)
        l.P("Tick {} from {}".format(cnt, thread_data[2]['thread']), color='d')
    return  
  
  thr1 = threading.Thread(target=t1, name='S_T1', daemon=True)
  thr2 = threading.Thread(target=t2, name='S_T2', daemon=True)
  thread_data[1]['thread'] = thr1
  thread_data[2]['thread'] = thr2
  
  l.P("Starting threads...")
  start_thread(1)
  start_thread(2)
  
  sleep(5)
                                         
  mon.log_status()
  
  
  l.P("Stopping thr1")
  stop_thread(1)
  l.P("Thread thr1 is_alive={}".format(thr1.is_alive()))
  l.P("Stopping thr2")
  stop_thread(2)
  l.P("Thread thr2 is_alive={}".format(thr2.is_alive()))
  
  mon.log_status()
  
  l.P("Sleeping another 2 sec", color='g')
  sleep(2)

  mon.log_status()
    
  # l.dict_pretty_format(mon.threads_data, display=True)

  # mon.display_process_tree_info()


import numpy as np
from collections import deque

from naeural_core.utils.sensors.virtual_time_series import VirtualTimeSeries

class BasicSeriesModel:
  """
  Basic timeseries prediction with constant bias (no variable step-id)
  """  
  def __init__(self, series_min=100, train_hist=None, train_periods=None):
    self.reset_params(
      train_hist=train_hist,
      series_min=series_min,
      train_periods=train_periods
    )
    return
  
  def reset_params(self, series_min=50, train_hist=None, train_periods=None):
    MIN_STEPS = 8
    assert series_min > MIN_STEPS, f"Minimal series config must be {MIN_STEPS + 1} steps. Config set to {series_min}"
    if train_hist is None:
      train_hist = int(series_min / (MIN_STEPS / 2))
    if train_periods is None:
      train_periods = series_min - (train_hist + 1)
    assert train_periods <= (series_min - train_hist)
    assert train_hist < train_periods
    self._min_series = series_min    
    self._train_steps = train_periods
    self._train_hist = train_hist
    return
  
  def fit(self, vect):
    # convert & add bias
    np_x = np.array(vect).reshape(-1)
    _min = np_x.min()
    _max = np_x.max()
    self._scale = _max - _min
    self._loc = _min
    np_x = (np_x - self._loc) / self._scale
    self._x_series = np_x
    l = np_x.shape[0]
    assert l >= self._min_series, "Minimal time-series len must be {}. Received {}".format(
      self._min_series, l)
    
    x_data = []
    y_data = []
    ts = self._train_steps
    th = self._train_hist
    for i in range(ts):
      start = -(th + i + 1)
      end = -(i+1)
      x_vect = np_x[start:end]
      y_val = np_x[end]
      x_data.append(x_vect)
      y_data.append(y_val)
    np_Xnb = np.array(x_data)
    np_b = np.ones((len(x_data), 1))
    np_X = np.concatenate((np_b, np_Xnb), axis=1)
    np_y = np.array(y_data).reshape((-1,1))
    self.theta = np.linalg.pinv(np_X.T.dot(np_X)).dot(np_X.T).dot(np_y)
    self.np_X = np_X
    self.np_y = np_y
    return
  
  def predict(self, nr_steps):
    x_data = deque(maxlen=self._train_hist)
    x_data.extend(self._x_series[-self._train_hist:])
    y_preds = []
    for i in range(nr_steps):
      np_x = np.array([1] + list(x_data))
      y_pred = np_x.dot(self.theta.ravel())
      y_preds.append(y_pred * self._scale + self._loc)
      x_data.append(y_pred.round(2))
    return y_preds
      
                       
    
if __name__ == '__main__':
  sensor = VirtualTimeSeries(n_steps=30, test_size=0.3, no_random=6)
  import matplotlib.pyplot as plt
  import seaborn as sns
  sns.set()
  
  np.set_printoptions(linewidth=200)
  np.set_printoptions(precision=1)
  
  train = sensor.get_train_data()
  xtrn = range(train.shape[-1])
  test = sensor._data_test
  xtst = range(train.shape[-1], train.shape[-1] + test.shape[-1])
  plt.figure(figsize=(26,8))
  plt.plot(xtrn, train.ravel(), c='black')
  plt.plot(xtst, test.ravel(), c='red')
  
  model = BasicSeriesModel(series_min=train.shape[-1])
  model.fit(train)
  yh = model.predict(nr_steps=test.ravel().shape[0])
  plt.plot(xtst, yh, c='green')
  
  
  plt.title(str(sensor._random_seed))
        


#local dependencies
from naeural_core import constants as ct

from naeural_core.utils.alerts import AlertHelper
    

class _AlerterMixin(object):
  def __init__(self):
    self.__alert_helpers = {}
    super(_AlerterMixin, self).__init__()
    return

  def _create_alert_state_machine(self):
    self.__alert_helpers = {}
    self.alerter_create()
    return
  

  @property
  def alerters_names(self):
    return list(self.__alert_helpers.keys())
  
  def alerter_maybe_create(self, alerter, **kwargs):
    if alerter not in self.__alert_helpers:
      self.alerter_create(alerter=alerter, **kwargs)
    return

  def alerter_create(self, alerter='default', 
                     raise_time=None, lower_time=None,
                     value_count=None,
                     raise_thr=None, lower_thr=None,
                     alert_mode=None,
                     alert_mode_lower=None,
                     reduce_value=None, reduce_threshold=None,
                     show_version=False):
    if alerter in self.__alert_helpers:
      raise ValueError("`AlertHelper` '{}' already exists".format(alerter))

    raise_time = self.cfg_alert_raise_confirmation_time if raise_time is None else raise_time
    lower_time = self.cfg_alert_lower_confirmation_time if lower_time is None else lower_time
    value_count = self.cfg_alert_data_count if value_count is None else value_count
    raise_thr = self.cfg_alert_raise_value if raise_thr is None else raise_thr
    lower_thr = self.cfg_alert_lower_value if lower_thr is None else lower_thr
    alert_mode = self.cfg_alert_mode if alert_mode is None else alert_mode
    alert_mode_lower = self.cfg_alert_mode_lower if alert_mode_lower is None else alert_mode_lower
    reduce_value = self.cfg_alert_reduce_value if reduce_value is None else reduce_value

    reduce_threshold = self._get_alerter_reduce_threshold() if reduce_threshold is None else reduce_threshold

    self.P("  Creating alerter '{}' for {}: data, thr_raise/time_raise, thr_lower/time_lower: {}, {}/{} , {}/{}".format(
      alerter,
      self._signature, value_count,
      raise_thr, raise_time, lower_thr, lower_time), color='y')

    if not (raise_time < lower_time):
      self.P("    WARNING: Alerter raise time should be below lowering time: raise={} vs lower={}".format(
        raise_time, lower_time), color='r')

    self.__alert_helpers[alerter] = AlertHelper(
      name=self._signature,
      values_count=value_count,
      raise_confirmation_time=raise_time,
      lower_confirmation_time=lower_time,
      raise_alert_value=raise_thr,
      lower_alert_value=lower_thr,
      alert_mode=alert_mode,
      alert_mode_lower=alert_mode_lower,
      reduce_value=reduce_value,
      reduce_threshold=reduce_threshold,
      show_version=show_version,
    )
    return
  
  def _check_alerter(self, alerter, maybe_create=True):
    if maybe_create:
      self.alerter_maybe_create(alerter)
    if alerter not in self.__alert_helpers:
      raise ValueError("`AlertHelper` '{}' does NOT exist".format(alerter))
    return
      

  def add_alerter_observation(self, value, alerter='default'):
    self._check_alerter(alerter)
    self.__alert_helpers[alerter].add_observation(
      value=value,
      # maybe threshold changes during night and we decided to "reduce" procs in the alerter
      # we can still use the alert_value but we need to reduce based on another threshold
      reduce_threshold=self._get_alerter_reduce_threshold(),
      )
    return


  
  def get_alerter(self, alerter='default'):
    # TODO: Either remove this one - no acces to actual object or put `alerter` as public
    self._check_alerter(alerter)
    return self.__alert_helpers[alerter]
  
  def get_alerter_status(self, alerter='default'):
    if alerter in self.__alert_helpers:
      return str(self.__alert_helpers[alerter])
    return
  
  
  def alerter_status_dict(self, alerter='default'):
    return dict(
      is_alert=self.alerter_is_alert(alerter=alerter),
      is_new_raise=self.alerter_is_new_raise(alerter=alerter),
      is_new_lower=self.alerter_is_new_lower(alerter=alerter),
      is_alert_new_raise=self.alerter_is_new_alert(alerter=alerter),
      is_alert_new_lower=self.alerter_is_new_lower(alerter=alerter),
      is_alert_status_changed=self.alerter_status_changed(alerter=alerter),
    )
 
  
  def alerter_get_last_value(self, alerter='default'):
    self._check_alerter(alerter)
    return self.__alert_helpers[alerter].get_last_raw_value()
  
  
  def alerter_in_confirmation(self, alerter='default'):
    self._check_alerter(alerter)
    return self.__alert_helpers[alerter].in_confirmation


  def alerter_setup_values(self, alerter='default'):
    self._check_alerter(alerter)
    return self.__alert_helpers[alerter].get_setup_values()


  def alerter_hard_reset(self, state=False, alerter='default'):
    self._check_alerter(alerter)
    self.P("WARNING: Alerter '{}' reset requested!".format(alerter), color='r')
    return self.__alert_helpers[alerter].hard_reset(default=state)
  
  
  def alerter_hard_reset_all(self):
    for alerter_name in self.__alert_helpers:
      self.alerter_hard_reset(alerter=alerter_name)
      
      
  # PUBLIC and DOCUMENTED methods
      
  def alerter_add_observation(self, value, alerter='default'):
    """
    Add a new numerical value observation to the given alerter state machine instance

    Parameters
    ----------
    value : float
      The value to be added that can be a percentage or a number or elements - depeding of the configuration of the alerter
      that has been given via "ALERT_MODE".
    alerter : str, optional
      The identifier of the given alerter state machine. The default is 'default'.

    Returns
    -------
    TYPE
      None.

    """
    return self.add_alerter_observation(value, alerter=alerter)     
  
  
  def alerter_is_new_alert(self, alerter='default'):
    """
    Returns `True` if the current state of the given `alerter` state machine has just changed from "lowered" to "raised"

    Parameters
    ----------
    alerter : str, optional
      Identifier of the given alerter instance. The default is 'default'.

    Returns
    -------
    TYPE
      bool.

    """
    self._check_alerter(alerter)
    return self.__alert_helpers[alerter].is_new_raise()


  def alerter_is_new_raise(self, alerter='default'):
    """
    Returns `True` if the current state of the given `alerter` state machine has just changed from "lowered" to "raised"

    Parameters
    ----------
    alerter : str, optional
      Identifier of the given alerter instance. The default is 'default'.

    Returns
    -------
    TYPE
      bool.

    """
    return self.alerter_is_new_alert(alerter=alerter)


  def alerter_is_new_lower(self, alerter='default'):
    """
    Returns `True` if the current state of the given `alerter` state machine has just changed from "raised" to "lowered"

    Parameters
    ----------
    alerter : str, optional
      Identifier of the given alerter instance. The default is 'default'.

    Returns
    -------
    TYPE
      bool.

    """
    self._check_alerter(alerter)
    return self.__alert_helpers[alerter].is_new_lower()


  def alerter_is_alert(self, alerter='default'):
    """
    Returns `True` if the current state of the given `alerter` state machine is "raised"

    Parameters
    ----------
    alerter : str, optional
      Identifier of the given alerter instance. The default is 'default'.

    Returns
    -------
    TYPE
      bool.

    """
    self._check_alerter(alerter)
    return self.__alert_helpers[alerter].is_alert()


  def alerter_status_changed(self, alerter='default'):
    """
    Returns `True` if the current state of the given `alerter` state machine has just changed

    Parameters
    ----------
    alerter : str, optional
      Identifier of the given alerter instance. The default is 'default'.

    Returns
    -------
    TYPE
      bool.

    """
    self._check_alerter(alerter)
    return self.__alert_helpers[alerter].status_changed()  
      
      
  def alerter_time_from_last_change(self, alerter='default'):
    """
    Returns the number of seconds from the last change of the given alerter state machine

    Parameters
    ----------
    alerter : str, optional
      Identifier of the given alerter instance. The default is 'default'.

    Returns
    -------
    TYPE
      bool.

    """
    self._check_alerter(alerter)
    return self.__alert_helpers[alerter].get_time_from_change()
  
  def alerter_maybe_force_lower(self, max_raised_time=0, alerter='default'):
    """
    Forces the given alerter to reset to "lowered" status if the current state is "raised"

    Parameters
    ----------
    alerter : str, optional
      Identifier of the given alerter instance. The default is 'default'.
      
    max_raised_time: float, optional
      The number of seconds after the raised alerter is forced to lower its status

    Returns
    -------
    TYPE
      bool.

    """
    was_reset = False
    self._check_alerter(alerter)
    if self.alerter_is_alert(alerter) and self.alerter_time_from_last_change(alerter) > max_raised_time:
      self.alerter_hard_reset(state=False, alerter=alerter)
      was_reset = True
    return was_reset
  
  def alerter_get_last_alert_duration(self, alerter='default'):
    """
    
    """
    self._check_alerter(alerter)
    alerter : AlertHelper = self.__alert_helpers[alerter]
    duration = alerter.get_last_alert_duration()
    if not alerter.is_new_lower():
      duration = None
    return duration
  


import abc
import os
import uuid
from naeural_core import DecentrAIObject

class BaseScoringPlugin(DecentrAIObject):

  def __init__(self, log, y_true_src, owner=None, **kwargs):
    self.y_true = None
    self.payload = None
    self.exceptions = None
    self.max_score = None
    self.score_per_obs = None
    self.config = None
    self.owner = owner

    self._y_true_src = y_true_src
    super(BaseScoringPlugin, self).__init__(log=log, **kwargs)
    return

  def startup(self):
    super().startup()
    self.y_true = self._get_y_true(self._y_true_src)
    return

  def _get_y_true(self, y_true_src):
    y_true = None
    delete_file = False
    file_path = None

    if isinstance(y_true_src, str):
      if os.path.isfile(y_true_src):
        file_path = y_true_src
      elif y_true_src.startswith('http'):
        saved_files, _ = self.log.maybe_download(
          url=y_true_src,
          fn='tmp-{}.json'.format(uuid.uuid4()),
          target='output'
        )
        file_path = saved_files[0]
        delete_file = True
      #endif
    else:
      y_true = y_true_src
    # endif

    if file_path is not None:
      try:
        y_true = self.log.load_json(fname=file_path, folder=None)
      except Exception as e:
        self.P("Could not load y_true json from {}:\n{}".format(file_path, e))

      if delete_file:
        os.remove(file_path)
    # endif
    return y_true

  @property
  def cfg_max_score(self):
    return self.config.get('MAX_SCORE', None)

  @property
  def cfg_score_per_obs(self):
    return self.config.get('SCORE_PER_OBS', 5)

  def _refresh(self, payload, config=None):
    self.payload = payload
    self.y_hat = self.payload["Y_HAT"]
    self.exceptions = []
    self.config = config or {}

    self.max_score = self.cfg_max_score

    if self.max_score is not None:
      self.score_per_obs = self.max_score / len(self.y_true)
    else:
      self.score_per_obs = self.cfg_score_per_obs
      self.max_score = self.score_per_obs * len(self.y_true)
    # endif
    return

  @abc.abstractmethod
  def _scoring(self):
    raise NotImplementedError()

  def _time_penalization(self, movie_seconds, processing_secondss):
    thr_1 = 0.02 * movie_seconds
    thr_2 = 0.1 * movie_seconds

    diff = abs(processing_secondss - movie_seconds)

    if diff < thr_1:
      return 0
    elif thr_1 <= diff < thr_2:
      return -5
    else:
      return -100


  def score(self, payload, config=None):
    self._refresh(payload=payload, config=config)

    try:
      s = self._scoring()
    except Exception as e:
      s = None
      msg = "Could not score due to the following error:\n{}".format(e)
      self.exceptions.append(msg)
      self.P(msg, color='r')

    movie_seconds = payload.get('MOVIE_SECONDS', None)
    processing_seconds = payload.get('PROCESSING_SECONDS', None)
    time_penalization = None

    if movie_seconds is not None and processing_seconds is not None:
      time_penalization = self._time_penalization(
        movie_seconds,
        processing_seconds
      )

    return s, time_penalization



import os
import json

from naeural_core import Logger

if __name__ == '__main__':
  l = Logger('KEYS', base_folder='.', app_folder='_local_cache')
  
  os.environ['EE_P1'] = 's1'
  os.environ['EE_P2'] = 's2'
  os.environ['EE_P3'] = 's3'  

  data = {
    '1' : [{
        'k1' : '$EE_P1',
        '2': 2,
      },
      {}
    ],
    'p2' : '$EE_P2',
    'p3' : '$EE_P3',
  }   
  
  l.save_data_json(data, 'no_secrets.txt')
  
  print(json.dumps(
    l.load_data_json('no_secrets.txt', replace_environment_secrets='$EE_'), 
      indent=4)
  )
  


from sklearn.datasets import make_blobs
from sklearn.model_selection import train_test_split
from naeural_core.utils.sensors.base import AbstractSensor

class VirtualSensor(AbstractSensor):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    return
  
  def _setup_connection(self, 
                        n_max_samples=1000, 
                        n_feats=2, 
                        centers=1, 
                        test_size=0.5,
                        no_random=False,
                        **kwargs,
                        ):
    random_state = 1234 if no_random else None
    data, centers = make_blobs(
      n_samples=n_max_samples,
      n_features=n_feats,
      centers=centers,
      return_centers=False,
      random_state=random_state,
      center_box=(-3,3)
      )
    x_train, x_test = train_test_split(data, test_size=test_size, random_state=random_state)
    self._data_train = x_train
    self._data_feed = x_test
    self._datastream_len = x_test.shape[0]
    return
  
  def _get_single_observation(self):
    data = self._data_feed[self._pos]
    self._pos += 1
    if self._pos >= self._datastream_len:
      self._pos = 0
    return data
  
    
  def _has_training_data(self):
    return True
    
  def _get_training_data(self):
    return self._data_train
  
  
    
  


# -*- coding: utf-8 -*-
import abc
import json
from collections import deque

from naeural_core import constants as ct
from naeural_core.data.base.base_plugin_dct import DataCaptureThread
from naeural_core.io_formatters.io_formatter_manager import IOFormatterManager
from naeural_core.comm import AMQPWrapper, MQTTWrapper

_CONFIG = {
  **DataCaptureThread.CONFIG,

  'CAP_RESOLUTION'  : 50, # overwrite default cap resolution - we should iterate faster on IoT data
  'LIVE_FEED'       : False,  
  
  'MAX_IDLE_TIME'   : 60,  
  
  'DEBUG_IOT_PAYLOADS' : False,

  'RECONNECTABLE': True,
  'ONE_AT_A_TIME': False,


  "HOST": '#DEFAULT',
  "PORT": '#DEFAULT',
  "USER": '#DEFAULT',
  "PASS": '#DEFAULT',
  "QOS": '#DEFAULT',
  "TOPIC": "#DEFAULT",
  "SECURED": "#DEFAULT",
  "PROTOCOL": "#DEFAULT",
  
  "MESSAGE_FILTER": {},
  "PATH_FILTER": [None, None, None, None],

  "URL": None,
  "STREAM_CONFIG_METADATA": {
    "HOST": '#DEFAULT',
    "PORT": '#DEFAULT',
    "USER": '#DEFAULT',
    "PASS": '#DEFAULT',
    "QOS": '#DEFAULT',
    "TOPIC": "#DEFAULT",
    "SECURED": "#DEFAULT",
    "PROTOCOL": "#DEFAULT",
  },

  'VALIDATION_RULES': {
    **DataCaptureThread.CONFIG['VALIDATION_RULES'],
  },
}


class BaseIoTQueueListenerDataCapture(DataCaptureThread):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(BaseIoTQueueListenerDataCapture, self).__init__(**kwargs)
    self.message_queue = deque(maxlen=1000)
    self.connected = False
    self.subscribed = False
    self._io_formatter_manager: IOFormatterManager = None
    return

  @property
  def _conn_type(self):
    """Property that defines the type of connection to create. Possible types: mqtt, amqp

    Returns:
        str: one of the following: 'mqtt', 'amqp'
    """
    ret = self.shmem['config_communication']['TYPE']
    if 'PROTOCOL' in self.cfg_stream_config_metadata and self.cfg_stream_config_metadata['PROTOCOL'] != '#DEFAULT':
      ret = self.cfg_stream_config_metadata['PROTOCOL'].lower()
    return ret

  def __get_stream_config_metadata_property(self, property):
    """Get the specific connection property from STREAM_CONFIG_METADATA. If it is not explicitly defined, consider the default value.
    The default value is the one specified in the communication layer.

    Args:
        property (str): the field representing a communication parameter

    Returns:
        str/int: the specified value defined by the user or the default value
    """
    params = self.shmem['config_communication']['PARAMS']
    ret = params.get(property)
    if self.config.get(property) != '#DEFAULT':
      ret = self.config.get(property)
    elif property in self.cfg_stream_config_metadata and self.cfg_stream_config_metadata.get(property) != '#DEFAULT':
      ret = self.cfg_stream_config_metadata.get(property)

    return ret

  def __get_topic(self):
    params = self.shmem['config_communication']['PARAMS']
    ret = params['PAYLOADS_CHANNEL']['TOPIC']
    if 'TOPIC' in self.cfg_stream_config_metadata and self.cfg_stream_config_metadata.get('TOPIC') != '#DEFAULT':
      ret = self.cfg_stream_config_metadata.get('TOPIC')
    return ret

  def _init(self):
    # use the parameters from the comm layer as default for this connection, if unspecified
    params = self.shmem['config_communication']['PARAMS']
    self._io_formatter_manager = self.shmem['io_formatter_manager']

    # build the config dict with all connection paramteres required by wrapper server
    self._comm_config = {
      ct.COMMS.EE_ID: params.get(ct.COMMS.EE_ID, None),
      "CUSTOM_CHANNEL": {
        "TOPIC": self.__get_topic()
      },
      ct.COMMS.HOST: self.__get_stream_config_metadata_property(ct.COMMS.HOST),
      ct.COMMS.PORT: self.__get_stream_config_metadata_property(ct.COMMS.PORT),
      ct.COMMS.USER: self.__get_stream_config_metadata_property(ct.COMMS.USER),
      ct.COMMS.PASS: self.__get_stream_config_metadata_property(ct.COMMS.PASS),
      ct.COMMS.QOS: self.__get_stream_config_metadata_property(ct.COMMS.QOS),
      ct.COMMS.SECURED: self.__get_stream_config_metadata_property(ct.COMMS.SECURED),
    }

    # build the kwargs of the wrapper server
    # TODO: maybe add low-level filtering of messages in the pub-sub wrapper
    #       this would allow for a more efficient message handling and no unwanted messages
    #       in the DCT message queue
    wrapper_kwargs = dict(
      log=self.log,
      config=self._comm_config,
      recv_channel_name="CUSTOM_CHANNEL",
      recv_buff=self.message_queue,
      connection_name=''.join([self._device_id, '_IoT_Listener_', self.cfg_name]),
    )

    # define the wrapper server
    if self._conn_type == 'mqtt':
      self.wrapper_server = MQTTWrapper(**wrapper_kwargs)
    elif self._conn_type == 'amqp':
      self.wrapper_server = AMQPWrapper(**wrapper_kwargs)
    else:
      raise ValueError("Cannot understand reduce controller type: {}".format(self.wrapper_server))

    self._maybe_reconnect_to_controller_server()
    return

  def _release(self):
    dct_ret = self.wrapper_server.release()
    for msg in dct_ret['msgs']:
      self.P(msg)
    self.connected = False
    self.subscribed = False
    del self.wrapper_server
    return

  def _maybe_reconnect_to_controller_server(self):
    """Connect to the server and send a notification with the result of the attempt.
    """
    if self.wrapper_server.connection is None or not self.connected:
      self.P("Trying to connect to the pub-sub server...")
      self.subscribed = False
      dct_ret = self.wrapper_server.server_connect()
      self.connected = dct_ret['has_connection']
      msg = dct_ret['msg']
      msg_type = dct_ret['msg_type']
      self.P("IoT DCT status post reconnect:\n{}".format(self.json_dumps(msg, indent=2)))
      self._create_notification(
        notif=msg_type,
        msg=msg
      )
    return

  def _maybe_reconnect(self):
    self._maybe_reconnect_to_controller_server()
    if not self.subscribed:
      self.P("Trying to subscribe to the pub-sub server...")
      if self._conn_type == 'amqp':
        dct_ret = self.wrapper_server.establish_one_way_connection('recv')
      elif self._conn_type == 'mqtt':
        dct_ret = self.wrapper_server.subscribe()
      else:
        dct_ret = None
      # endif

      self.P("IoT DCT status post subscribe:\n{}".format(self.json_dumps(dct_ret, indent=2)))
      self.subscribed = dct_ret['has_connection']
      msg = dct_ret['msg']
      msg_type = dct_ret['msg_type']
      self._create_notification(
        notif=msg_type,
        msg="IoTDCT Status:" + msg
      )
    # endif

    return

  def _maybe_fill_message_queue(self):
    """Call the receive method associated with the controller server, which can fill the buffer with messages

    Raises:
        e: Exception from receiving, induces by a connection issue
    """
    try:
      self.wrapper_server.receive()
    except Exception as e:
      self.connected = False
      self.subscribed = False
      self.P(str(e), color='r')
      raise e
    # end try-except
    return


  def __extract_and_process_one_message(self):
    """
    This method extracts one message from the message queue and processes it via __process_iot_message
    """
    msg = self.message_queue.popleft()
    dict_msg = json.loads(msg)
    processed_message, message_type = self.__process_iot_message(dict_msg)

    if processed_message is None:
      return
    
    if self.cfg_debug_iot_payloads:
      _path = processed_message.get(self.ct.PAYLOAD_DATA.EE_PAYLOAD_PATH, [None, None, None, None])
      self.P(f"Accepted message of type {message_type} from {_path}")
    
    if message_type == "struct_data":
      self._add_struct_data_input(processed_message)
    elif message_type == "image":
      self._add_img_input(processed_message)
    else:
      self.P("Unknown message type: {}".format(message_type), color='r')
      self.P("Full message: {}".format(processed_message), color='r')
    return


  def _extract_and_process_messages(self, nr_messages=1):
    """ This method extracts and processes a number of messages from the message queue"""
    for _ in range(nr_messages):
      self.__extract_and_process_one_message()
    return


  def _run_data_aquisition_step(self):
    if len(self.message_queue) == 0:
      return
    
    if self.cfg_one_at_a_time:
      nr_messages = 1
    else:
      nr_messages = min(len(self.message_queue), self.cfg_stream_window)

    self._extract_and_process_messages(nr_messages)
    return


  def __process_iot_message(self, msg):
    """Decode the message if it is in a format supported by an Execution Engine, and then parse and filter it to support custom logic.
    The former is useful when there are multiple Execution Engines in a network and all send messages with different formatters.

    Parameters:
    ----------
        msg (dict): The raw message received by the listener

    Returns:
    ----------    
        dict/img: The message that will be sent downstream, maybe formatted, parsed and filtered
    """
    result = msg
    message_type = "unknown"

    formatter = self._io_formatter_manager.get_required_formatter_from_payload(result)
    if formatter is not None:
      result = formatter.decode_output(result)
    else:
      # we kind of treat this case already, because the default formatter is considered the identity function
      result = msg
    # endif formatter is not None

    result = self.__filter_message(result)
    if result is not None:
      result = self._parse_message(result)
    # endif result is not None

    # TODO: maybe add support for numpy arrays as struct data
    if result is None:
      message_type = "ignored_message"
    elif isinstance(result, self.np.ndarray) or isinstance(result, self.PIL.Image.Image):
      message_type = "image"
    elif isinstance(result, dict) or isinstance(result, list) or isinstance(result, tuple) or isinstance(result, str):
      message_type = "struct_data"
    # endif decide message type

    return result, message_type
  
  
  def __filter_message_by_path(self, unfiltered_message):
    """Filter messages that get passed forward using the path filter

    Parameters:
    ----------
        unfiltered_message (dict): message received from the queue server, possibly formatted if it was in a format supported by the Execution Engine

    Returns:
    ----------
        dict: the message that satisfies certain conditions or None if it does not satisfy them
    """
    path_filter = self.cfg_path_filter
    result = unfiltered_message
    if isinstance(unfiltered_message, dict):
      path = unfiltered_message.get(self.ct.PAYLOAD_DATA.EE_PAYLOAD_PATH, [None, None, None, None])
      path = [x.upper() if isinstance(x, str) else x for x in path]
      for i in range(4):
        _path_filter =  path_filter[i]
        if _path_filter is not None:
          if not isinstance(_path_filter, list):
            _path_filter = [_path_filter]
          _path_filter = [x.upper() if isinstance(x, str) else x for x in _path_filter]
          if path[i] not in _path_filter:
            if self.cfg_debug_iot_payloads:
              self.P(f"Path filter {path_filter} dropped {path}")
            result = None
            break
    return result



  def __filter_message(self, unfiltered_message):
    """Filter messages that get passed forward

    Args:
        unfiltered_message (dict): message received from the queue server, possibly formatted if it was in a format supported by the Execution Engine

    Returns:
        dict: the message that satisfies certain conditions or None if it does not satisfy them
    """
    # first filter by path
    result = self.__filter_message_by_path(unfiltered_message)
    if result is not None:
      # then filter by message filter dict
      dct_filter = self.cfg_message_filter
      if dct_filter is not None:
        is_valid = self.dict_in_dict(dct_filter, result)
      if is_valid:
        result = self._filter_message(result)
    return result


  @abc.abstractmethod
  def _filter_message(self, unfiltered_message):
    """
    Overwrite this method to filter messages that get passed forward

    Parameters
    ----------
    unfiltered_message : dict
        message received from the queue server, possibly formatted if it was in a format supported by the Execution Engine

    Returns
    -------
    filtered_message : dict | None
        the message that satisfies certain conditions or none if it does not satisfy them
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def _parse_message(self, filtered_message):
    """
    Overwrite this method to parse messages that get passed forward

    Parameters
    ----------
    filtered_message : dict
        message received from the queue server, possibly formatted if it was in a format supported by the Execution Engine

    Returns
    -------
    parsed_message : dict | tuple | str | np.ndarray | PIL.Image.Image
        The message that will be sent downstream, either as an image or as a struct data
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def _parse_and_filter_message(self, message):
    """Overwrite this method to parse and filter messages that get passed forward

    Args:
        messages (dict): message received from the queue server, possibly formatted if it was in a format suported by the Execution Engine

    Returns:
        dict/img: the message that satisfies certain conditions
    """
    raise NotImplementedError()



# TODO Bleo: WIP
from typing import List, Tuple, Any
import torch as th

import numpy as np
from naeural_core.local_libraries.nn.th.training.callbacks.base import TrainingCallbacks


class AddWhiteNoise:
  def __init__(self, prc_noise):
    self.prc_noise = prc_noise
    return

  def __call__(self, img : th.tensor):
    b,c,h,w = img.shape
    l = b*h*w
    mask = th.tensor([False for _ in range(l)])
    np_indexes = np.random.choice(np.arange(l), size=int(self.prc_noise*l))
    mask[np_indexes] = True
    mask = mask.reshape((b,h,w)).to(img.device)
    w = th.where(mask)
    img = img.transpose(1,2).transpose(2,3)
    img[w] = 255
    img = img.transpose(2,3).transpose(1,2)
    return img


class AutoencoderTrainingCallbacks(TrainingCallbacks):

  def __init__(self, prc_noise=0.3, **kwargs):
    self._prc_noise = prc_noise
    super(AutoencoderTrainingCallbacks, self).__init__(**kwargs)
    return

  def _lst_augmentations(self, **kwargs) -> List[Tuple[Any, dict]]:
    lst_transformations = []
    if self._prc_noise != 0:
      noise_transformation = (AddWhiteNoise, dict(prc_noise=self._prc_noise))
      lst_transformations.append(noise_transformation)
    #endif
    return lst_transformations

  def _get_y(self, lst_y):
    y = np.vstack(lst_y)
    return y

  def _get_y_hat(self, lst_y_hat):
    y_hat = np.vstack(lst_y_hat)
    return y_hat

  def _evaluate_callback(self, epoch: int, dataset_info: dict, y: np.ndarray, y_hat: np.ndarray, idx: np.ndarray, key: str = 'dev') -> dict:
    return {'{}_loss'.format(key): self._owner.average_loss}

  def _test_callback(self, dataset_info: dict, y: np.ndarray, y_hat: np.ndarray, idx: np.ndarray) -> dict:
    return {}



import os
import math
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
from PIL import Image
from enum import Enum
from abc import abstractmethod
from ratio1 import BaseDecentrAIObject


__version__ = '1.0.1.1'
__VER__ = __version__


class VisEngineEnum(Enum):
  CV2 = 0
  PIL = 1

class RGBColorEnum(Enum):
  DEFAULT_BLACK     = (0, 0, 0)
  DEFAULT_WHITE     = (255, 255, 255)
  DEFAULT_RED       = (255, 0, 0)
  DEFAULT_GREEN     = (0, 255, 0)
  DEFAULT_BLUE      = (0, 0, 255)
  
  DARK_BLUE    = (11, 45, 109)
  LIGHT_BLUE   = (129, 255, 254)
  ORANGE       = (250, 149, 19)
  YELLOW       = (250, 222, 98)
  RED          = (248, 70, 100)
  GREEN        = (50, 200, 50)

IMG_EXT = ['.png', '.bmp', '.jpg', '.jpeg']

class BaseVisualization(BaseDecentrAIObject):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    return
  
  @abstractmethod
  def display(self, img, window_name):
    raise NotImplementedError()
  
  @abstractmethod
  def close_display(self):
    raise NotImplementedError()
  
  @abstractmethod  
  def draw_rect(self, img, top, left, bottom, right, color, thickness):
    raise NotImplementedError()
  
  @abstractmethod
  def draw_box(self, img, top, left, bottom, right, display_str, color, thickness):
    raise NotImplementedError()
  
  def draw_boxes(self, img, lst_boxes, lst_display_str, lst_color, thickness):
    raise NotImplementedError()
    
  @abstractmethod
  def draw_text(self, img, text, left, top, color, thickness, font_scale):
    raise NotImplementedError() 
  
  @abstractmethod
  def read_image(self, path):
    raise NotImplementedError()
    
  @abstractmethod
  def save_image(self, path, img):
    raise NotImplementedError()
    
  def resize_image(self, height, width, interpolation):
    raise NotImplementedError()
  
  def is_image(self, path):
    return any(path.endswith(x) for x in IMG_EXT)
  
  def read_images(self, fn):
    if isinstance(fn, str) and os.path.isdir(fn):
      lst_files = [x for x in os.listdir(fn) if self.is_image(x)]
      lst_paths = [os.path.join(fn, x) for x in lst_files]
    elif isinstance(fn, str):
      if os.path.exists(fn) and self.is_image(fn):
        lst_paths = [fn]
      else:
        raise ValueError('File named incorrect: {}'.format(fn))
    elif isinstance(fn, list):
      lst_paths = [x for x in fn if self.is_image(x)]
    else:
      raise ValueError('Value not understood: {}'.format(fn))
    lst_bgr = [self.read_image(x) for x in tqdm(lst_paths)]
    lst_rgb = [self.reverse_channels(x) for x in lst_bgr]
    return lst_paths, lst_bgr, lst_rgb

  def save_images(self, lst_imgs, reverse_channels, path=None):
    self.log.p('Saving images')
    if path is None:
      path = self.log.now_str()
    os.makedirs(path, exist_ok=True)
    for idx, img in tqdm(enumerate(lst_imgs)):
      if reverse_channels:
        img = self.reverse_channels(img)
      self.save_image(os.path.join(path, '{}.png'.format(idx)), img)
    self.log.p('Done saving images', show_time=True)
    return

  def show(self, img, title=None, xlabel=None, ylabel=None):
    plt.imshow(img)
    if title:
      plt.title(title)
    if xlabel:
      plt.xlabel(xlabel)
    if ylabel:
      plt.ylabel(ylabel)
    plt.show()
    return

  def side_by_side(self, lst_imgs, lst_titles=None, lst_xlabel=None, lst_ylabel=None, title=None):
    for img in lst_imgs:
      if img.max() <= 1: assert img.dtype in ['float32', 'float64']
      if img.max() > 1 : assert img.dtype == 'uint8'
    
    if lst_titles:
      assert len(lst_imgs) == len(lst_titles)
    if lst_xlabel:
      assert len(lst_imgs) == len(lst_xlabel)
    if lst_ylabel:
      assert len(lst_imgs) == len(lst_ylabel)
    
    nr = len(lst_imgs)
    for idx, img in enumerate(lst_imgs):
      plt.subplot(1, nr, idx+1)
      if lst_titles: plt.title(lst_titles[idx])
      if lst_xlabel: plt.xlabel(lst_xlabel[idx])
      if lst_ylabel: plt.ylabel(lst_ylabel[idx])
      plt.imshow(img)
    if title:
      plt.title(title)
    plt.show()
    return
  
  def display_images(self, imgs, n_cols, 
                     titles=None, nr_max=20, 
                     axis_off=False, title=None,
                     figsize=(13,8)):
    if titles:
      assert len(imgs) == len(titles)
    
    if nr_max is not None:
      imgs = imgs[:nr_max]    
    n_rows = int(math.ceil(len(imgs) / n_cols))      
    crt_img = 0
    fig, axs = plt.subplots(n_rows, n_cols, figsize=figsize)
    for row in range(n_rows):
      for col in range(n_cols):
        idx = row * n_cols + col
        if idx >= len(imgs):
          img = np.zeros(shape=(10, 10, 3), dtype=np.uint8)
        else:
          img = imgs[idx]
        if len(axs.shape) == 1:
          axs[col].imshow(img)
        else:
          axs[row, col].imshow(img)
        if titles:
          lbl = ''
          if idx < len(titles):
            lbl = titles[idx]
          if len(axs.shape) == 1:
            axs[col].set_title(lbl)
          else:
            axs[row, col].set_title(lbl)
        if axis_off:
          axs[row, col].axis('off')
        crt_img+= 1
        if crt_img >= (n_rows * n_cols):
          break
      #endfor
      if crt_img >= (n_rows * n_cols):
        break
    if title:
      fig.suptitle(title)
    plt.show()    
    return
  
  def display_in_batches(self, imgs, n_cols, batch_size=16, titles=None, title=None):
    if titles:
      assert len(imgs) == len(titles)
    
    nr_batches = int(math.ceil(len(imgs) / batch_size))
    for nr in tqdm(range(nr_batches)):
      start = batch_size * nr
      stop = start + batch_size if nr < nr_batches - 1 else len(imgs)
      x_batch = imgs[start:stop]
      if titles:
        y_batch = titles[start:stop]
      else:
        y_batch = None
      plot_title = '{} batch {}'.format(title, nr + 1) if title else 'Batch {}'.format(nr + 1)
      self.display_images(x_batch, n_cols=n_cols, titles=y_batch, title=plot_title)
    return
  
  def reverse_channels(self, img):
    img = img[:, :, ::-1]
    return img
      
  def transform_image(self, img, normalize, reverse_channels):
    if normalize:
      img/= 255
    if reverse_channels:
      img = self.reverse_channels(img)
    return img  
  
  def reverse_color(self, color):
    return color[::-1]
  
  def is_image(self, img_path):
    return any([img_path.endswith(x) for x in IMG_EXT])
  
  def plot_distribution(self, classes):
    assert isinstance(classes, (list, np.ndarray))
    np_unique, np_counts = np.unique(classes, return_counts=True)
    plt.bar(np.arange(len(np_unique)), np_counts)
    plt.xticks(np_unique)
    plt.title('Class distribution')
    plt.xlabel('Class')
    plt.ylabel('Nr occurance')
    plt.show()
    return
  
  
  
class VisualizationCV2(BaseVisualization):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    import cv2
    self.__cv2 = cv2
    self.font = kwargs.get('font', self.__cv2.FONT_HERSHEY_COMPLEX)
    return
  
  def close_display(self):
    self.__cv2.waitKey(1)
    for i in range(5):
      self.__cv2.waitKey(1)
      self.__cv2.destroyAllWindows()
    return
  
  def show(self, img, window_name='Debug'):
    if isinstance(img, str):
      img = self.read_image(img)
    self.__cv2.imshow(window_name, img)
    return
  
  def display(self, img, window_name='Debug'):
    self.show(img, window_name)
    self.__cv2.waitKey(0)
    self.close_display()
    return
  
  def read_image(self, path):
    assert os.path.exists(path)
    return self.__cv2.imread(path)
  
  def save_image(self, path, img):
    self.__cv2.imwrite(path, img)
    return
  
  def resize_image(self, img, height, width, interpolation=None):
    if interpolation is None:
      interpolation = self.__cv2.INTER_LINEAR
    dim = (width, height)
    img_res = self.__cv2.resize(img, dim, interpolation=interpolation)
    return img_res
  
  def draw_text(self, img, text, left, top, color=RGBColorEnum.DEFAULT_BLACK.value, thickness=1, font_scale=1):
    self.__cv2.putText(img=img, text=text, org=(left, top), color=color, thickness=thickness, 
                fontScale=font_scale, fontFace=self.font)
    return
  
  def draw_rect(self, img, top, left, bottom, right, color=RGBColorEnum.DEFAULT_GREEN.value, thickness=2):
    self.__cv2.rectangle(img=img, pt1=(left, top), pt2=(right, bottom), color=color, thickness=thickness)
    return
  
  def draw_box(self, img, top, left, bottom, right, display_str='', color=RGBColorEnum.DEFAULT_GREEN.value, thickness=2):
    #bounding box
    self.draw_rect(img=img, top=top, left=left, bottom=bottom, right=right, color=color, thickness=thickness)
    
    if display_str != '':
      font_scale = 0.6
      font_thickness = 1
      (tw, th) = self.__cv2.getTextSize(text=display_str, fontFace=self.font, fontScale=font_scale, thickness=font_thickness)[0]
      top_desc = top - th
      bottom_desc = top
      left_desc = left
      right_desc = left_desc + tw
      self.draw_rect(img=img, top=top_desc, left=left_desc, bottom=bottom_desc, right=right_desc, color=color, thickness=self.__cv2.FILLED)
      self.draw_text(img=img, text=display_str, left=left_desc, top=bottom_desc, color=RGBColorEnum.DEFAULT_BLACK.value,
                     thickness=font_thickness, font_scale=font_scale)
    #endif
    return
  
  def draw_boxes(self, img, lst_boxes, lst_display_str, lst_color=None, thickness=2):
    if lst_color is None:
      lst_color = [tuple([int(x) for x in np.random.randint(low=0, high=255, size=3)]) for _ in range(len(lst_boxes))]
    
    assert len(lst_boxes) == len(lst_display_str) == len(lst_color)        
    for box, display_str, color in zip(lst_boxes, lst_display_str, lst_color):
      top, left, bottom, right = box
      self.draw_box(img=img, top=top, left=left, bottom=bottom, right=right,
                    display_str=display_str, color=color, thickness=thickness)
    return img
  
  def draw_inferences(self, img, lst_preds, lst_color=None):
    lst_boxes = [x['TLBR_POS'] for x in lst_preds]
    lst_display_str = ['{}: {:.2f}%'.format(x['TYPE'], x['PROB_PRC']) for x in lst_preds]
    self.draw_boxes(
      img=img,
      lst_boxes=lst_boxes,
      lst_display_str=lst_display_str,
      lst_color=lst_color
      )
    return img
  
class VisualizationPIL(BaseVisualization):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    return
  
  def close_display(self):
    self.log.p('PIL opens images in separate processes. Please parse process and close the opend image')
    return
  
  def show(self, img, window_name='Debug'):
    if isinstance(img, str):
      img = self.read_image(img)
    if isinstance(img, np.ndarray):
      img = Image.fromarray(img)
    img.show()
    return
  
  def display(self, img, window_name='Debug'):
    self.show(img, window_name)
    return
  
  def read_image(self, path, return_numpy=True):
    assert os.path.exists(path)
    img = Image.open(path)
    if return_numpy:
      img = np.asarray(img)
    return img
  
  def save_image(self, path, img):
    if isinstance(img, np.ndarray):
      img = Image.fromarray(img)
    elif isinstance(img, Image):
      pass
    else:
      raise ValueError('Please provide either np image or PIL Image')
    img.save(path)
    return
  
  def resize_image(self, img, height, width, interpolation=Image.BILINEAR):
    size = (width, height)
    if isinstance(img, np.ndarray):
      img = Image.fromarray(img)
    img_res = img.resize(size=size, resample=interpolation)
    img_res = np.asarray(img_res)
    return img_res
  
  
class VisEngine(BaseDecentrAIObject):
  def __init__(self, vis_engine=VisEngineEnum.CV2, **kwargs):
    super().__init__(**kwargs)    
    self._init_eng(vis_engine, **kwargs)
    return
  
  def _init_eng(self, vis_engine, **kwargs):
    if vis_engine == VisEngineEnum.CV2:
      self.vis_eng = VisualizationCV2(**kwargs)
    else:
      self.vis_eng = VisualizationPIL(**kwargs)
    return
  
  def get_eng(self):
    return self.vis_eng

if __name__ == '__main__':
  from naeural_core import Logger
  log = Logger(lib_name='TST', config_file='config.txt')
  
  vis = VisEngine(log=log, vis_engine=VisEngineEnum.PIL).get_eng()
  img = vis.read_image('C:/Users/ETA/Dropbox/DATA/_vapor_data/_cervical_lessions/_data/train/0/20150722014/20150722163342.jpg')
  vis.show(img)
  vis.save_image('a.png', img)
  
  
  
  
  
  
  


import cv2 as cv
import cProfile


VIDEO_PATHS = [
  '/home/bleo/debug/1',
  '/home/bleo/debug/2',
  '/home/bleo/debug/3',
  '/home/bleo/debug/4',
  # '/home/bleo/debug/telefon_1.mp4',
  # '/home/bleo/debug/telefon_2.mp4',
  # '/home/bleo/debug/telefon_3.mp4',
  # '/home/bleo/debug/telefon_4.mp4',
  # '/home/bleo/debug/car demo.mp4',
  # '/home/bleo/debug/Test_2.mp4',
  # '/home/bleo/debug/Test 2 - Portret 1 - 608p.mp4',
  # '/home/bleo/debug/Test 2 - Portret 2 - 144p.mp4',
  # '/home/bleo/debug/Test 2 - Portret 3 - 240p.mp4',
  # '/home/bleo/debug/Test 2 - Portret 4 - 240p.mp4',
  # '/home/bleo/debug/Test 2 - Portret 5 - 608p.mp4'
]

N_LAPS = 100


def get_rotate_code_from_orientation(cap_orientation, for_undo=False):
  rotate_code = None
  if int(cap_orientation) == 270:
    rotate_code = cv.ROTATE_90_CLOCKWISE
  elif int(cap_orientation) == 180:
    rotate_code = cv.ROTATE_180
  elif int(cap_orientation) == 90:
    rotate_code = cv.ROTATE_90_COUNTERCLOCKWISE

  if rotate_code is not None and for_undo:
    rotate_code = 2 - rotate_code
  return rotate_code


def correct_rotation(frame, rotateCode):
  if rotateCode is None:
    return frame
  return cv.rotate(frame, rotateCode)


def original_video_reader(video_path, to_show=False):
  cap = cv.VideoCapture(video_path)

  while cap.isOpened():
    ret, frame = cap.read()
    if ret == True:
      if to_show:
        cap_orientation = cap.get(cv.CAP_PROP_ORIENTATION_META)
        cv.imshow(f'OriginalReader_{cap_orientation}', frame)
        pressed_key = cv.waitKey(25) & 0xFF
        # Press Q on keyboard to  exit
        if pressed_key == ord('q'):
          break
    # Break the loop
    else:
      break

  # When everything done, release the video capture object
  cap.release()
  # Closes all the frames
  cv.destroyAllWindows()
  return


def good_rotation_video_reader(video_path, to_show=False):
  cap = cv.VideoCapture(video_path)
  cap.set(cv.CAP_PROP_ORIENTATION_AUTO, 0)
  cap_orientation = cap.get(cv.CAP_PROP_ORIENTATION_META)
  rotate_code = get_rotate_code_from_orientation(cap_orientation)

  while cap.isOpened():
    ret, frame = cap.read()
    if ret == True:
      if rotate_code is not None:
        frame = correct_rotation(frame, rotate_code)

      if to_show:
        cv.imshow(f'UpdatedReader_{cap_orientation}', frame)
        pressed_key = cv.waitKey(25) & 0xFF
        # Press Q on keyboard to  exit
        if pressed_key == ord('q'):
          break
    # Break the loop
    else:
      break
  # When everything done, release the video capture object
  cap.release()
  # Closes all the frames
  cv.destroyAllWindows()
  return


def check_video(video_path):
  original_video_reader(video_path=video_path, to_show=True)
  good_rotation_video_reader(video_path=video_path, to_show=True)

  return

def main1():
  for _ in range(N_LAPS):
    for path in VIDEO_PATHS:
      original_video_reader(path)
  return


def main2():
  for _ in range(N_LAPS):
    for path in VIDEO_PATHS:
      good_rotation_video_reader(path)
  return


def main(video_path):
  cap = cv.VideoCapture(video_path)
  cap.set(cv.CAP_PROP_ORIENTATION_AUTO, 0)
  cap_orientation = cap.get(cv.CAP_PROP_ORIENTATION_META)
  rotate_code = get_rotate_code_from_orientation(cap_orientation)

  sz = int(cap.get(cv.CAP_PROP_FRAME_COUNT))
  center_frame_id = int(sz // 2)
  back_frame_id = int(sz * 0.1)
  forward_frame_id = int(sz * 0.9)


  while cap.isOpened():
    ret, frame = cap.read()
    if ret == True:

      if rotate_code is not None:
        frame = correct_rotation(frame, rotate_code)
      # Display the resulting frame
      cv.imshow('Video', frame)

      pressed_key = cv.waitKey(25) & 0xFF
      # Press Q on keyboard to  exit
      if pressed_key == ord('q'):
        break
      elif pressed_key == ord('c'):
        cap.set(cv.CAP_PROP_POS_FRAMES, center_frame_id-1)
      elif pressed_key == ord('b'):
        cap.set(cv.CAP_PROP_POS_FRAMES, back_frame_id-1)
      elif pressed_key == ord('f'):
        cap.set(cv.CAP_PROP_POS_FRAMES, forward_frame_id-1)

    # Break the loop
    else:
      break

  # When everything done, release the video capture object
  cap.release()

  # Closes all the frames
  cv.destroyAllWindows()


if __name__ == '__main__':
  # video_paths = [
  #   '/home/bleo/debug/car demo.mp4',
  #   '/home/bleo/debug/Test_2.mp4',
  #   '/home/bleo/debug/Test 2 - Portret 1 - 608p.mp4',
  #   '/home/bleo/debug/Test 2 - Portret 2 - 144p.mp4'
  # ]
  # video_path = '/home/bleo/debug/car demo.mp4'
  # video_path = '/home/bleo/debug/Test_2.mp4'
  # video_path = '/home/bleo/debug/Test 2 - Portret 1 - 608p.mp4'
  # video_path = '/home/bleo/debug/Test 2 - Portret 2 - 144p.mp4'
  #
  # dps = 25
  # main(video_path=video_path)
  # main1()
  # cProfile.run('main1()')
  # cProfile.run('main2()')

  for path in VIDEO_PATHS:
    check_video(path)





import numpy as np

from naeural_core.utils.sanchez_utils.Actor import *
from naeural_core.utils.sanchez_utils import utils


class Environment:
  def __init__(self, name='', h=90, w=170, actors=None, seconds_per_turn=1, background=None):
    self._last_map = None
    self.name = name
    self.h = h
    self.w = w
    self.background = background if background is not None else ct.WHITE
    self.seconds_per_turn = seconds_per_turn
    self.running = False
    self.turns_passed = 0

    self._map = np.zeros((h, w), np.int32)
    self._last_map = self._map
    # self._frame = np.full((h, w, 3), self.background, np.uint8)

    self._stones = []
    self.actors = []
    if actors is not None:
      for actor in actors:
        self.add_actor(actor)

    self.img = None

  def add_stone(self, x, y):
    if 0 < x < self.h and 0 < y < self.w:
      if self._map[x, y] > 0:
        actor = self.actors[self._map[x, y] - 1]
        if actor.get_species() == TYPE_SNAKE:
          actor.clear_map(self._map)
        actor.retire()
      self._map[x, y] = -1
      self._stones.append((x, y))
    return

  def _remove_stone(self, x, y):
    if self._in_map(x, y):
      if self._map[x, y] == -1:
        self._map[x, y] = 0
        stones_set = set(self._stones)
        if (x, y) in stones_set:
          stones_set.remove((x, y))
          self._stones = list(stones_set)
    return

  def _reset_stones(self):
    for (x, y) in self._stones:
      self._remove_stone(x, y)
    return

  def update_stones(self, stones):
    stones_set = set(self._stones)
    for (x, y) in stones:
      if (x, y) not in stones_set:
        self.add_stone(x, y)
    return

  def get_stones(self):
    return self._stones

  def add_actor(self, actor):
    x, y = actor.matrix_position(self.h, self.w)
    if self._map[x, y] < 0:
      # a stone already there
      return

    self.actors.append(actor)
    actor_id = len(self.actors)
    self.actors[len(self.actors) - 1].id = actor_id
    self._map[x, y] = actor_id

  def get_actor(self, idx=1):
    idx = idx - 1
    if 0 <= idx < len(self.actors):
      return self.actors[idx]
    return None

  def get_actors_data(self):
    # used for displaying actors names
    datas = []
    for actor in self.actors:
      text = actor.get_data_string()
      species = actor.__class__.__name__.lower()
      x, y = actor.matrix_position(self.h, self.w)
      a = actor.alpha
      b = actor.beta
      points = actor.get_points()
      id = actor.get_id()
      datas.append((text, species, x, y, a, b, points, id))

    return datas

  def get_env_colors(self):
    # color[0] will be the color of the rocks
    # color[1] will be the color of the background
    # color[i + 1] will be the color of the actor
    # to get the colored map use colors[env.get_map() + 1]
    colors = np.full((len(self.actors) + 2, 3), self.background, np.uint8)
    colors[0] = utils.negate_color(self.background)

    for actor in self.actors:
      colors[actor.get_id() + 1] = actor.get_color()

    return colors

  def _check_collisions(self):
    N = len(self.actors)
    for i in range(N):
      for j in range(i + 1, N):
        if self.actors[i].check_collision(self.actors[j]):
          self.actors[i].change_direction()
          self.actors[j].change_direction()

  def _reset_map(self):
    for actor in self.actors:
      # xmin, xmax, ymin, ymax = actor.get_coords()
      # self._map[xmin: xmax, ymin: ymax] = 0
      self._map[actor.matrix_position(self.h, self.w)] = 0

  def _in_map(self, x, y):
    return -1 < x < self.h and -1 < y < self.w

  def _map_position(self, point):
    x, y = point
    x, y = max(x, 0), max(y, 0)
    h, w = self._map.shape
    x, y = min(x, h - 1), min(y, w - 1)
    point = x, y
    return point

  def _update_map(self):
    for actor in self.actors:
      points = actor.get_points()
      for point in points:
        self._map[self._map_position(point)] = actor.get_id()

    self._last_map = self._map

  def get_current_map(self):
    return self._map

  def get_map(self):
    return self._last_map

  def get_frame(self, scale_x=1, scale_y=1):
    big_map = self._map.repeat([scale_x] * self.w, axis=1)
    big_map = big_map.repeat([scale_y] * self.h, axis=0)

    colors = np.full((len(self.actors) + 2, 3), self.background, np.uint8)
    colors[0] = utils.negate_color(self.background)

    for actor in self.actors:
      colors[actor.id + 1] = actor.color

    return colors[big_map + 1]

  def stop(self, verbose=False):
    self.running = False
    if verbose:
      print('Env should stop')

  def dead_environment(self):
    for actor in self.actors:
      if actor.life_state != RETIRED_STATE:
        return False
    return True

  def take_turn(self, verbose=False):
    if verbose:
      print('Iteration on Environment loop')
    self._reset_map()

    for actor in self.actors:
      actor.forward(self.h, self.w, self)

    # self._check_collisions()

    self._update_map()
    self.turns_passed += 1
    # time.sleep(self.seconds_per_turn)
    if verbose:
      print(f'Done turn #{self.turns_passed}')

  def start(self, turn_limit=None, verbose=False, auto_turn=False):
    self._update_map()
    if verbose:
      print(f'Turn limit is {turn_limit}')

    self.running = True

    if auto_turn:
      while self.running and (turn_limit is None or self.turns_passed < turn_limit):
        self.take_turn(verbose)

      if verbose:
        print(f'Environment {self.name} stopped after {self.turns_passed} turns.')

    # endif
    return












# global dependencies
import numpy as np
import os
import pandas as pd

# local dependencies
from ratio1 import load_dotenv
from naeural_core import constants as ct
from naeural_core import Logger
from naeural_core.serving.model_testing.base import Base
from naeural_core.serving.model_testing.test_all.data import files_dict as TEST_FILES


SORT_BY_RAW = True


EXCEPTION_FILES = [
  '__init__.py',
]


EXCEPTION_SERVINGS = [
  # 'th_cqc',
  'th_iqa_fast', 'th_iqa_slow',
  'th_yf_xp', 'th_rface_id',
  'th_yf_base', 'code_llama_v2',
  'th_effdet7', 'th_plano_s1',
  'th_yf5l', 'th_yf5s',
]


class GenericServingTester(Base):
  def draw_inference_particular_template(self, img, inference, **kwargs):
    return img

  def draw_inference_default(self, img, inference, **kwargs):
    # First, we get the label
    label = kwargs.get('label', None)
    if label is None:
      inf_type = inference.get(ct.TYPE, None) or ''
      inf_prc = inference.get(ct.PROB_PRC, None) or 0
      label = f'{inf_type} | {round(float(inf_prc), 2)}'
    # endif label is None
    # Then, we draw the inference
    top, left, bottom, right = inference[ct.TLBR_POS]
    color = kwargs.get('color', ct.GREEN)
    img = self._painter.draw_detection_box(
      image=img,
      top=top,
      left=left,
      bottom=bottom,
      right=right,
      label=label,
      prc=round(float(inference[ct.PROB_PRC]), 2),
      color=color,
      font_scale=1
    )
    return img

  def draw_inference(self, img, inference, **kwargs):
    if self.get_model_name() in []:
      return self.draw_inference_particular_template(img, inference, **kwargs)
    return self.draw_inference_default(img, inference, **kwargs)

  def plot(self, dataset_name, **kwargs):
    lst_imgs = self.get_images(dataset_name)
    preds = self.get_last_preds()
    for i in range(len(lst_imgs)):
      img = lst_imgs[i][:, :, ::-1]
      img = np.ascontiguousarray(img)
      curr_preds = preds['INFERENCES'][0][i]

      for pred in curr_preds:
        img = self.draw_inference(
          img=img,
          inference=pred,
          is_valid=True,
          color=ct.GREEN
        )

      if self._show_plots:
        self._painter.show(self._model_name, img)
      if self._save_plots:
        self._painter.save(
          img,
          fn='{}/{}/{}.jpg'.format(
            self._testing_subfolder_path,
            dataset_name,
            f'{i + 1}'
          )
        )
      # endif self._save_plots
    # endfor
    return


def compress_dict(nested_dict, key):
  res = []
  for id, lst_files in nested_dict.items():
    for fn in lst_files:
      res.append(fn)
    # endofr lst_files
  # endfor nested_dict.items()
  return {
    key: res
  }


if __name__ == '__main__':
  import multiprocessing as mp
  mp.set_start_method('spawn')
  log = Logger('MTA', base_folder='.', app_folder='_local_cache', TF_KERAS=False)
  GPU_DEVICE = 0

  # check if the serving files exist
  servings_dir = 'core/serving/default_inference'
  if not os.path.exists(servings_dir):
    log.P(f'ERROR: {servings_dir} does not exist')
    log.P(f'Exiting...')
    exit(1)
  # endif not os.path.exists

  # get the serving files
  servings = os.listdir(servings_dir)
  # retrieve the serving names
  serving_names = [
    s[:-3] for s in servings if s.endswith('.py') and s not in EXCEPTION_FILES
  ]
  # define the used data and the tests for each serving
  serving_data = {
    k: {
      'data': TEST_FILES,
      'tests': [{}]  # this will run the serving with its default configuration
    }
    for k in serving_names if k not in EXCEPTION_SERVINGS
  }

  total_df = pd.DataFrame()
  save_subdir = os.path.join('testing', f'{log.file_prefix}_TEST_ALL')

  for i, (serving_name, test_dict) in enumerate(serving_data.items()):
    current_tests = test_dict['tests']
    current_data = test_dict['data']
    log.P(f'[({i + 1} / {len(serving_data)})]Running tests for serving {serving_name} with test configs:\n{current_tests}')
    try:
      test_process = GenericServingTester(
        log=log,
        model_name=serving_name,
        test_files=current_data,
        gpu_device=GPU_DEVICE,
        save_plots=False,
        show_plots=False,
        nr_warmup=0,
        nr_predicts=1,
        inprocess=False,
        print_errors=True,
        label_extension='txt'
      )

      load_dotenv()
      dct_params = {
        "WITNESS_INFO": True,
        "NEW_PARAM": 0,
        'USE_FP16': False,
        'MAX_WAIT_TIME': 20,
        'DEFAULT_DEVICE': 'cuda:' + str(GPU_DEVICE),
        "MODEL_ZOO_CONFIG": {
          "endpoint": os.environ["EE_MINIO_ENDPOINT"],
          "access_key": os.environ["EE_MINIO_ACCESS_KEY"],
          "secret_key": os.environ["EE_MINIO_SECRET_KEY"],
          "secure": os.environ["EE_MINIO_SECURE"],
          "bucket_name": "model-zoo"
        }
      }
      current_df = test_process.run_tests(
        lst_tests=current_tests,
        dct_params=dct_params
      )
      total_df = pd.concat([total_df, current_df])
      log.save_dataframe(total_df, 'results.csv', folder='output', subfolder_path=save_subdir)
      log.P(f'[({i + 1} / {len(serving_data)})]Successfully done tests for serving {serving_name}')
    except Exception as e:
      log.P(f'[({i + 1} / {len(serving_data)})]Failed tests for serving {serving_name} with test configs:\n{current_tests}')
  # endfor serving_data




# TODO Bleo: WIP
import torch.nn as nn
from naeural_core.local_libraries.nn.th.encoders import SimpleImageEncoder, SimpleImageDecoder, BASIC_ENCODER

_CONFIG = {
  'GRID_SEARCH' : {
    'GRID' : {
      'in_channels' : [3],
      'prc_noise' : [
        0.3,
        0.5
      ],
      'image_size': [
        # [32, 32],
        # [128, 256],
        [448, 640]
      ],
      'layers': [None]
    },
    'CALLBACKS_PARAMS': ['prc_noise'],
    'DATA_PARAMS': ['image_size'],
    'MODEL_PARAMS': ['in_channels', 'image_size', 'layers'],
  }
}


class AutoencoderModelFactory(nn.Module):

  def __init__(self, in_channels=3, image_size=None, layers=None, **kwargs):
    super().__init__()
    layers = layers or BASIC_ENCODER
    image_size = image_size or [64, 64]
    self.in_channels = in_channels
    self.image_size = image_size

    self.encoder = SimpleImageEncoder(
      h=self.image_size[0],
      w=self.image_size[1],
      channels=self.in_channels,
      layers=layers
    )
    self.decoder = nn.Sequential(
      SimpleImageDecoder(
        h=self.image_size[0],
        w=self.image_size[1],
        channels=self.in_channels,
        layers=layers
      ),
      nn.Sigmoid()
    )
    return

  def forward(self, x):
    x = self.encoder(x)
    x = self.decoder(x)
    return x




from time import time, sleep
import functools

def timer(func):
  @functools.wraps(func) # for debugging purposes; if not specified, foo.__name__ will be "wrapper" instead of "foo" (i.e. preserve original name, docstring, etc)
  def wrapper(*arg, **kwargs):
    start = time()
    res = func(*arg, **kwargs)
    return res, time()-start

  return wrapper

@timer
def foo(a,b,c=3):
  print(a,b,c)
  sleep(0.1)
  return

if __name__ == '__main__':

  x = foo("dfd","dfs")
  print("method foo returned '{}' (exec time: {:.2f}s)".format(x[0], x[1]))



from naeural_core.business.base import BasePluginExecutor

__VER__ = '1.0.0'

_CONFIG = {
  **BasePluginExecutor.CONFIG,

  'ALLOW_EMPTY_INPUTS': True,

  'VALIDATION_RULES': {
    **BasePluginExecutor.CONFIG['VALIDATION_RULES'],
  },
}


class PluginMonitor01Plugin(BasePluginExecutor):
  CONFIG = _CONFIG

  def on_init(self):
    self.__capture_manager = self.global_shmem["capture_manager"]
    self.__serving_manager = self.global_shmem["serving_manager"]
    self.__business_manager = self.global_shmem["business_manager"]

    self.requests_queue = []
    return

  def __send_error_payload(self, error_msg, data):
    self.P(error_msg, color='r')
    self.add_payload_by_fields(
      plugin_default_configuration=None,
      command_params=data,
      error=error_msg,
    )
    return

  def __check_plugin_type_ok(self, plugin_type):
    type_not_none = plugin_type is not None
    type_string = isinstance(plugin_type, str)
    type_correct = type_string and plugin_type in ["capture", "serving", "business"]

    type_ok = type_not_none and type_string and type_correct
    error_msg = None

    if not type_ok:
      error_msg = f"Plugin type incorrect, expected string, one of {['capture', 'serving', 'business']}, received {plugin_type}."
    # end if

    return type_ok, error_msg

  def __check_plugin_signature_ok(self, plugin_signature):
    signature_not_none = plugin_signature is not None
    signature_string = isinstance(plugin_signature, str)

    signature_ok = signature_not_none and signature_string
    error_msg = None

    if not signature_ok:
      error_msg = f"Plugin signature incorrect, expected string, received {plugin_signature}."
    # end if

    return signature_ok, error_msg

  def __get_plugin_default_config(self, plugin_type, plugin_signature):
    plugin_default_config = None
    if plugin_type.lower() == "business":
      plugin_default_config = self.__business_manager.get_plugin_default_config(plugin_signature)
    elif plugin_type.lower() == "capture":
      plugin_default_config = self.__capture_manager.get_plugin_default_config(plugin_signature)
    elif plugin_type.lower() == "serving":
      plugin_default_config = self.__serving_manager.get_plugin_default_config(plugin_signature)

    return plugin_default_config

  def on_command(self, data, plugin_default_configuration=None, plugin_type=None, plugin_signature=None, **kwargs):
    """
    Called when the instance receives new INSTANCE_COMMAND

    Parameters
    ----------
    data : any
      object, string, etc.

    Returns
    -------
    None.

    """

    if plugin_default_configuration is None or not plugin_default_configuration:
      self.P(f"WARNING! Received UNKNOWN command {str(data)}", color='r')
      return

    self.P("Received \"PLUGIN_DEFAULT_CONFIGURATION\" command...")

    plugin_type_ok, error_message = self.__check_plugin_type_ok(plugin_type)
    if not plugin_type_ok:
      self.__send_error_payload(error_message, data)
      return

    plugin_signature_ok, error_message = self.__check_plugin_signature_ok(plugin_signature)
    if not plugin_signature_ok:
      self.__send_error_payload(error_message, data)
      return

    self.requests_queue.append((plugin_type, plugin_signature, data))
    return

  def process(self):
    while len(self.requests_queue) > 0:

      plugin_type, plugin_signature, data = self.requests_queue.pop(0)

      plugin_default_config = self.__get_plugin_default_config(plugin_type, plugin_signature)

      if plugin_default_config is None:
        self.__send_error_payload(f"Plugin signature {plugin_signature} not found in {plugin_type} manager.", data)
        return
      else:
        self.add_payload_by_fields(
          plugin_default_configuration=plugin_default_config,
          command_params=data,
        )

    return



from naeural_core.manager import Manager
from naeural_core import constants as ct


class ScoringManager(Manager):

  def __init__(self, log, owner, **kwargs):
    self._dct_scorers = None
    self.owner = owner
    super(ScoringManager, self).__init__(log=log, prefix_log='[SCOM]', **kwargs)
    return

  def startup(self):
    super().startup()
    self._dct_scorers = self._dct_subalterns
    return

  def _get_plugin_class(self, name):
    _module_name, _class_name, _class_def, _class_config = self._get_module_name_and_class(
      locations=ct.PLUGIN_SEARCH.LOC_TESTING_FRAMEWORK_SCORING_PLUGINS,
      name=name,
      suffix=ct.PLUGIN_SEARCH.SUFFIX_TESTING_FRAMEWORK_SCORING_PLUGINS,
      safety_check=True,  # perform safety check
    )
    return _class_def

  def maybe_create_scorer(self, name, y_true_src):
    if y_true_src is None:
      return

    identification = name + '_' + self.log.hash_object(y_true_src, size=4)
    if identification in self._dct_scorers:
      return self._dct_scorers[identification]

    _cls = self._get_plugin_class(name)

    scorer = _cls(
      log=self.log,
      y_true_src=y_true_src,
      owner=self.owner
    )

    self._dct_scorers[identification] = scorer
    return scorer

  def score(self, y_true_src, payload, config):
    name = payload['TESTER_NAME']
    scorer_obj = self.maybe_create_scorer(
      name=name,
      y_true_src=y_true_src,
    )

    if scorer_obj is None:
      return

    s, time_penalization = scorer_obj.score(payload, config)

    if type(s) == dict:
      total_score = s['ACCURACY'] * scorer_obj.max_score + time_penalization
    else:
      total_score = s + time_penalization

    dct_score = {
      'FUNCTIONALITY_SCORE': s,
      'TIME_PENALIZATION': time_penalization,
      'TOTAL_SCORE': total_score,
      'MAX_SCORE': scorer_obj.max_score
    }

    return dct_score

  def get_y_true(self, y_true_src, name):
    scorer_obj = self.maybe_create_scorer(
      name=name,
      y_true_src=y_true_src,
    )

    if scorer_obj is None:
      return

    return scorer_obj.y_true


if __name__ == '__main__':
  from naeural_core import Logger
  log = Logger(
    lib_name='EE',
    base_folder='.',
    app_folder='_local_cache',
    config_file='config_startup.txt',
    max_lines=3000,
    TF_KERAS=False
  )
  manager = ScoringManager(log=log, owner=None)
  locations = ['plugins.serving.pipelines.architectures']
  signature = 'custom'
  suffix = 'ModelFactory'
  tmp = manager._get_module_name_and_class(
    locations=locations,
    name=signature,
    suffix=suffix,
    safety_check=False,  # perform safety check
  )



import cv2


from naeural_core.utils.basic_qr_reader import read_qr
  

if __name__ == '__main__':
  
  fns = [
    # 'xperimental/qr_codes/test1.png',
    # 'xperimental/qr_codes/test2.png',
    'xperimental/qr_codes/test3.png',
    'xperimental/qr_codes/pic1.jpg',
    'xperimental/qr_codes/pic2.jpg',
    ]
  
  for fn in fns:
    np_rgb_img = cv2.imread(fn)[:,:,::-1]
    cv2.imshow("img", np_rgb_img[:,:,::-1])
    cv2.waitKey(0)
    cv2.destroyAllWindows()
      
    ###
  
    res = read_qr(np_rgb_img) 
    
    if res is not None:
      data, tlbr, bbox, str_qr, np_slice = res
      
      if bbox is not None:
        print(f"QRCode for {fn}:\n{data}")
        # display the image with lines
        # length of bounding box
        n_lines = len(bbox)
        for i in range(n_lines):
          # draw all lines
          point1 = tuple(bbox[i][0].astype('int').tolist())
          point2 = tuple(bbox[(i+1) % n_lines][0].astype('int').tolist())
          cv2.line(np_rgb_img, point1, point2, color=0, thickness=5)  
        
      
      cv2.imshow("slice", np_slice)
      cv2.waitKey(0)
      cv2.destroyAllWindows()


from naeural_core.data.base.base_iot_queue_listener import \
    BaseIoTQueueListenerDataCapture

_CONFIG = {
  **BaseIoTQueueListenerDataCapture.CONFIG,

  'VALIDATION_RULES': {
    **BaseIoTQueueListenerDataCapture.CONFIG['VALIDATION_RULES'],
  },
}


class IoTQueueListenerImageDataCapture(BaseIoTQueueListenerDataCapture):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(IoTQueueListenerImageDataCapture, self).__init__(**kwargs)
    return

  def _filter_message(self, unfiltered_message):
    filtered_message = None
    if isinstance(unfiltered_message, str):
      # We received a string, we assume it is a base64 encoded image (but we should check this)
      if self.log.base64_to_np_image(unfiltered_message) is not None:
        filtered_message = unfiltered_message
      # end if image is valid

    elif isinstance(unfiltered_message, dict):
      if 'IMG' in unfiltered_message and isinstance(unfiltered_message['IMG'], str):
        if self.log.base64_to_np_image(unfiltered_message['IMG']) is not None:
          filtered_message = unfiltered_message
        # end if image is valid
      # end if image is in dict
    # end if message is string or dict

    return filtered_message

  def _parse_message(self, filtered_message):
    # We want to return a numpy image, as we send images downstream

    # we already checked that the message is valid
    image = None
    if isinstance(filtered_message, str):
      image = self.log.base64_to_np_image(filtered_message)
    if isinstance(filtered_message, dict):
      image = self.log.base64_to_np_image(filtered_message['IMG'])
    return image



"""
TODO: Modify stream window for more than 1 data step after plugins such as net-config support multiple data steps.

2024-11-16: This is a single data step DCT designed for plugins that expect only one data step from upstream.

"""

from naeural_core.data.default.iot.iot_queue_listener import IoTQueueListenerDataCapture


_CONFIG = {
  **IoTQueueListenerDataCapture.CONFIG,
  
  'MAX_DEQUE_LEN'   : 32, 
  'STREAM_WINDOW'   : 1,
  'ONE_AT_A_TIME'   : True,
  
  'DEBUG_IOT_PAYLOADS' : False,
  
  


  'VALIDATION_RULES': {
    **IoTQueueListenerDataCapture.CONFIG['VALIDATION_RULES'],
  },
}


class NetworkListenerDataCapture(IoTQueueListenerDataCapture):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(NetworkListenerDataCapture, self).__init__(**kwargs)
    return

  def _init(self):
    super(NetworkListenerDataCapture, self)._init()
    self.P(f"Initializing {self.__class__.__name__} with filter {self.cfg_path_filter} and message filter {self.cfg_message_filter}")
    return



from datetime import datetime
from copy import copy

class _SqlQueryAcquisitionMixin(object):
  def __init__(self):
    self.stored_procedures = None
    self.queries = None
    super(_SqlQueryAcquisitionMixin, self).__init__()
    return

  def _get_sql_queries(self, current_datetime):
    """
    Returns queries to be run at the current timestep - based on the current timestep and the past
      ran queries.

    :param current_datetime: datetime
    :return: dict(DATASOURCE_NAME: query string)
    """
    sql_queries = {}

    if self.stored_procedures is not None:
      for i in range(len(self.stored_procedures)):
        if self._has_new_data(self.stored_procedures[i], current_datetime):
          last_query_datetime = self.stored_procedures[i].get(
            'LAST_QUERY_DATETIME',
            self._backlog_start_time
          )

          sql_queries[self.stored_procedures[i]['DATASOURCE_NAME']] = self._parse_sp(
              self.stored_procedures[i], current_datetime, last_query_datetime
          )
          self.stored_procedures[i]['LAST_QUERY_DATETIME'] = datetime.strftime(
            current_datetime,
            "%Y-%m-%d %H:%M:%S"
          )
        # endif
      # endfor

    if self.queries is not None:
      for i in range(len(self.queries)):
        if self._has_new_data(self.queries[i], current_datetime):
          last_query_datetime = self.queries[i].get(
            'LAST_QUERY_DATETIME',
            self._backlog_start_time
          )

          sql_queries[self.queries[i]['DATASOURCE_NAME']] = self._parse_query(
            self.queries[i],
            current_datetime,
            last_query_datetime
          )

          self.queries[i]['LAST_QUERY_DATETIME'] = datetime.strftime(
            current_datetime,
            "%Y-%m-%d %H:%M:%S"
          )
        # endif
      # endfor
    # endif
    return sql_queries

  def _save_state(self):
    """
    Saves last ran datetime for each query and stored procedure

    :return:
    """
    if not self.cfg_save_last_state:
      return

    obj = {
      'stored_procedures': {},
      'queries': {}
    }

    if self.stored_procedures is not None:
      obj['stored_procedures'] = {
        sp['DATASOURCE_NAME']: sp for sp in self.stored_procedures
      }

    if self.queries is not None:
      obj['queries'] = {
          q['DATASOURCE_NAME']: q for q in self.queries
      }

    self.persistence_serialization_save(obj)
    return

  def _load_state(self):
    """
    Loads last ran datetime for each query and stored procedure

    :return:
    """
    obj = self.persistence_serialization_load()
    if obj is None:
      return

    for name, details in obj['stored_procedures'].items():
      for i in range(len(self.stored_procedures)):
        if self.stored_procedures[i]['DATASOURCE_NAME'] == name:
            self.stored_procedures[i]['LAST_QUERY_DATETIME'] = details['LAST_QUERY_DATETIME']
        # endif
      # endfor
    # endfor

    for name, details in obj['queries'].items():
      for i in range(len(self.stored_procedures)):
        if self.queries[i]['DATASOURCE_NAME'] == name:
          self.queries[i]['LAST_QUERY_DATETIME'] = details['LAST_QUERY_DATETIME']
        # endif
      # endfor
    # endfor
    return

  def _has_new_data(self, dct_config, current_datetime):
    """
    Checks if a query needs to be ran at the current timestamp - based on its config,
      current datetime and last run time

    :param dct_config:
    :param current_datetime:
    :return:
    """
    period = dct_config.get("PERIOD", "DAILY")  ### Default period is daily
    last_query_time = datetime.strptime(
      dct_config.get("LAST_QUERY_DATETIME", self._backlog_start_time),
      "%Y-%m-%d %H:%M:%S")

    if last_query_time == datetime(1900, 1, 1, 0, 0):
      return True

    if period == 'DAILY':
      current_date = copy(current_datetime).replace(minute=0, hour=0)
      query_hour = dct_config.get("SCHEDULE_HOUR", None)
      if (current_date - last_query_time).days >= 1:
        if query_hour is not None:
          if query_hour == current_datetime.hour:
            return True
        else:
          return True
      return False
    elif period == 'HOURLY':
      if (current_datetime - last_query_time).seconds >= 3600 or \
              (current_datetime - last_query_time).days > 1:
        return True
      return False
    elif period == 'WEEKLY':
      current_date = copy(current_datetime).replace(minute=0, hour=0)
      query_hour = dct_config.get("SCHEDULE_HOUR", None)
      if (current_date - last_query_time).days >= 7:
        if query_hour is not None:
          if query_hour == current_datetime.hour:
            return True
        else:
          return True
      return False
    elif period == 'MINUTE':
      if (current_datetime - last_query_time).seconds >= 60 or \
              (current_datetime - last_query_time).days > 1:
        return True
      return False

    else:
      raise ValueError("Invalid configured query period `{}`".format(period))
    return


  def _parse_query(self, dct_query, current_datetime, last_query_datetime):
    """
    Method used to convert a query dict to the string required to be run for the current time window

    :param dct_query:
    :param current_datetime:
    :param last_query_datetime:
    :return:
    """
    query_string = "SELECT {} FROM {} WHERE {}".format(', '.join(
      dct_query['SELECT']),
      dct_query["TABLE"],
      self._get_filter_string(dct_query, current_datetime, last_query_datetime)
    )

    if dct_query.get("GROUP_BY", None) not in ["NONE", "None", None]:
      query_string += " GROUP BY {}".format(', '.join(dct_query['GROUP_BY']))

    return query_string

  def _get_filter_string(self, dct_query, current_datetime, last_query_datetime):
    """
    Method used to parse the 'WHERE' part of the query.

    :param dct_query:
    :param current_datetime:
    :param last_query_datetime:
    :return:
    """
    filter_string = "{} < CONVERT(DATETIME, '{}') ".format(
      dct_query['DATETIME_COLUMN'],
      current_datetime
    )
    filter_string += " AND {} > CONVERT(DATETIME, '{}') ".format(
      dct_query['DATETIME_COLUMN'],
      last_query_datetime,
    )

    if dct_query.get("FILTER", None) not in ["NONE", 'None', None]:
      filter_string += " AND {}".format(dct_query['FILTER'])

    return filter_string

  def _parse_sp(self, dct_stored_procedure, current_datetime, last_query_datetime):
    """
    Method used to create the query string needed to execute a stored procedure on the current
      time windows

    :param dct_stored_procedure: stored procedure config
    :param current_datetime: current datetime
    :param last_query_datetime: datetime of the last run (used for the start point of the window)

    :return: query string
    """
    query_string = "EXEC {} @{} = '{}', @{} = '{}'".format(
        dct_stored_procedure["SP_NAME"],
        dct_stored_procedure["START_DATE_PARAMETER_NAME"],
        last_query_datetime,
        dct_stored_procedure["END_DATE_PARAMETER_NAME"],
        current_datetime
      )
    return query_string





#global dependencies
import tensorflow as tf

#local dependencies
from .abstract_model import AbstractModel

class Tfodapi2Model(AbstractModel):
  def __init__(self, config_model, **kwargs):
    self.config_model = config_model
    super().__init__(**kwargs)
    return
  
  def _load_graph(self):
    self.log.start_timer('load_graph')    
    graph_file = self.config_model['GRAPH']
    graph = tf.saved_model.load(graph_file)
    if graph is None:
      raise ValueError("Graph loading failed for {}".format(graph_file))
    self.graph = graph
    self.log.stop_timer('load_graph')
    return
  
  def _process_input(self, inputs, **kwargs):
    return inputs
  
  def _predict(self, inputs, **kwargs):
    self.log.start_timer('predict')
    preds = self.graph.predict(inputs)
    self.log.stop_timer('predict')
    return preds  
  
  def init(self, **kwargs):
    return
  
  def load(self, **kwargs):
    self._load_graph()    
    return  
  
  def prepare(self, **kwargs):
    return
  
  
  
  
  
  
  
  
  
  
  


import cv2
from time import perf_counter
import numpy as np
from time import sleep
import os

if __name__ == '__main__':

  path = '/Users/laurentiupiciu/Dropbox/DATA/_vapor_data/_config_repo/LAURENTIU_STREAMS/testing_framework/aggressive_close_+_fallen.mp4'
  # path = '/home/laurentiu/Dropbox/DATA/_vapor_data/_config_repo/LAURENTIU_STREAMS/testing_framework/aggressive_close_+_fallen.mp4'

  cap = cv2.VideoCapture(path)

  lst_processing_times = []
  lst_sleep_times = []
  lst_iter_times = []
  frame_count = 0
  CAP_RESOLUTION = 25

  from naeural_core import Logger
  log = Logger(lib_name='DCTR', base_folder='.', app_folder='_local_cache', TF_KERAS=False)


  if False:
    t0 = perf_counter()
    while True:
      start_iter_time = perf_counter()
      start_processing_time = perf_counter()
      has_frame, frame = cap.read()
      if has_frame:
        frame_count += 1
        frame = np.ascontiguousarray(frame[:, :, ::-1])
      else:
        break
      processing_time = perf_counter() - start_processing_time
      lst_processing_times.append(processing_time)
      sleep_time = max(1/CAP_RESOLUTION - processing_time, 0.00001)
      lst_sleep_times.append(sleep_time)
      sleep(sleep_time)
      iter_time = perf_counter() - start_iter_time
      lst_iter_times.append(iter_time)
    #endwhile

    t1 = perf_counter()

    print("Total capture time for {} frames: {:.2f}s".format(frame_count, t1-t0))

  if False:
    CAPTURE_BATCH = 32
    t0 = perf_counter()
    while True:
      start_iter_time = perf_counter()
      start_processing_time = perf_counter()
      lst_raw_frames = []
      for _ in range(CAPTURE_BATCH):
        log.start_timer('cv2_capread')
        has_frame, frame = cap.read()
        log.end_timer('cv2_capread')
        if has_frame:
          frame_count += 1
          lst_raw_frames.append(frame)
        else:
          break
        #endif
      #endfor

      if len(lst_raw_frames) == 0:
        break

      log.start_timer('np.array')
      np_raw_frames = np.array(lst_raw_frames)
      log.end_timer('np.array')

      log.start_timer('channels')
      np_switched_raw_frames = np_raw_frames[:,:,:, ::-1]
      log.end_timer('channels')

      log.start_timer('ascontiguousarray')
      np_switched_raw_frames = np.ascontiguousarray(np_switched_raw_frames)
      log.end_timer('ascontiguousarray')

      log.start_timer('virtual_add_inputs')
      lst_proc_frames = [x for x in np_switched_raw_frames]
      log.end_timer('virtual_add_inputs')

      processing_time = perf_counter() - start_processing_time
      lst_processing_times.append(processing_time)
      sleep_time = max(1 / CAP_RESOLUTION - processing_time, 0.00001)
      lst_sleep_times.append(sleep_time)
      sleep(sleep_time)
      iter_time = perf_counter() - start_iter_time
      lst_iter_times.append(iter_time)
    # endwhile

    t1 = perf_counter()

    print("Total capture time for {} frames (batch={}): {:.2f}s".format(frame_count, CAPTURE_BATCH, t1 - t0))

    log.show_timers()

  if True:
    lst_raw_imgs = []
    for i in range(32):
      log.start_timer('cv2_capread')
      has_frame, frame = cap.read()
      log.end_timer('cv2_capread')
      assert has_frame
      lst_raw_imgs.append(frame)
    #endfor

    np_imgs = np.array(lst_raw_imgs)

    np_one_image = np_imgs[0].copy()

    for _ in range(20):
      log.start_timer('channels_one')
      x_c = np_one_image[:,:,::-1]
      log.end_timer('channels_one')

      log.start_timer('ascontiguousarray_one')
      x = np.ascontiguousarray(x_c)
      log.end_timer('ascontiguousarray_one')

    for _ in range(20):
      log.start_timer('channels_{}'.format(np_imgs.shape[0]))
      x_c = np_imgs[:,:,:,::-1]
      log.end_timer('channels_{}'.format(np_imgs.shape[0]))

      log.start_timer('ascontiguousarray_{}'.format(np_imgs.shape[0]))
      x = np.ascontiguousarray(x_c)
      log.end_timer('ascontiguousarray_{}'.format(np_imgs.shape[0]))

    log.show_timers()

